# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from abc import ABCMeta
from abc import abstractmethod

import xarray
import xarray.core.indexing as indexing

from earthkit.data.utils import ensure_dict
from earthkit.data.utils import ensure_iterable

from .attrs import AttrList
from .profile import Profile

LOG = logging.getLogger(__name__)

# These backend_kwargs are also direct xarray.open_dataset kwargs
BACKEND_AND_XR_OPEN_DS_KWARGS = ["decode_times", "decode_timedelta", "drop_variables"]

# These kwargs cannot be passed to xarray.open_dataset (not even inside backend_kwargs)
NON_XR_OPEN_DS_KWARGS = ["split_dims", "direct_backend"]


class VariableBuilder:
    def __init__(
        self,
        name,
        var_dims,
        data_maker,
        tensor,
        remapping,
        local_attr_keys=None,
        fixed_local_attrs=None,
    ):
        """
        Create a builder for a single variable in the dataset.

        Parameters
        ----------
        name: str
            The name of the variable.
        var_dims: list
            The dimensions of the variable.
        data_maker: callable
            A function that generates the data object stored in the xarray variable.
        tensor: Tensor
            The tensor object that contains the data.
        remapping: Remapping
            The remapping object that contains the remapping information.
        local_attr_keys: list, optional
            A list of metadata keys that are used to collect local attributes. These attributes
            are not taken into account by the attribute policies.
        fixed_local_attrs: dict, optional
            A dictionary of fixed local attributes that are added to the variable. These attributes
            are not taken into account by the attribute policies.
        """
        self.name = name
        self.var_dims = var_dims
        self.data_maker = data_maker
        self._attrs = {}
        self.local_keys = ensure_iterable(local_attr_keys)
        self.fixed_local_attrs = ensure_dict(fixed_local_attrs)
        self.tensor = tensor
        self.remapping = remapping

    def build(self, add_earthkit_attrs=True):
        if add_earthkit_attrs:
            if hasattr(self.tensor.source[0], "handle"):
                md = self.tensor.source[0].metadata().override()
                attrs = {
                    "message": md._handle.get_buffer(),
                    "bitsPerValue": md.get("bitsPerValue", 0),
                }
                self._attrs["_earthkit"] = attrs

        self._attrs.update(self.fixed_local_attrs)
        data = self.data_maker(self.tensor, self.var_dims, self.name)
        return xarray.Variable(self.var_dims, data, attrs=self._attrs)

    def collect_attrs(self, attrs, strict=True, extra_attrs=None):
        """Load the attributes for the variable.

        Parameters
        ----------
        attrs: list
            A list of attributes to be collected.
        strict: bool, optional
            If True, perform a strict check on the attributes.
        collect: list, optional
            A list of attributes that are collected but not added to the variable attributes.
        """
        attrs = attrs + AttrList(self.local_keys)

        res = {}
        keys_strict = []
        fixed_attrs = {}

        first = None

        def _metadata():
            nonlocal first
            return self.tensor.source[0].metadata() if not first else first

        for a in attrs:
            if a.name not in self.var_dims and a.name not in res and a.name not in self.fixed_local_attrs:
                if a.fixed():
                    fixed_attrs[a.name] = a.value()
                elif callable(a):
                    res.update(a(_metadata()))
                else:
                    if strict:
                        keys_strict.append(a.name)
                    else:
                        res.update(a.get(_metadata()))

        res = {k: v for k, v in res.items() if v is not None}

        # TODO: do we need a strict mode here? The extra cost has to be justified
        if keys_strict:
            assert strict
            v, _ = self.tensor.source.unique_values(keys_strict)
            res.update(v)

        self._attrs = res
        self._attrs.update(fixed_attrs)

        # extra attributes to be collected but not added to the variable
        collected_attrs = {}
        if extra_attrs:
            res = {}
            fixed_attrs = {}
            for a in extra_attrs:
                if a.fixed():
                    fixed_attrs[a.name] = a.value()
                elif callable(a):
                    res.update(a(_metadata()))
                else:
                    res.update(a.get(_metadata()))

            res = {k: v for k, v in res.items() if v is not None}
            collected_attrs = res
            collected_attrs.update(fixed_attrs)

        return {k: v for k, v in self._attrs.items() if k not in self.local_keys}, collected_attrs

    def adjust_attrs(self, drop_keys=None, rename=None):
        drop_keys = ensure_iterable(drop_keys)
        drop_keys = [k for k in drop_keys if k not in self.local_keys]
        r = {}
        for k, v in self._attrs.items():
            if k not in drop_keys:
                if isinstance(v, list) and len(v) == 1:
                    r[k] = v[0]
                else:
                    r[k] = v
        self._attrs = r
        # self._attrs = {k: v[0] for k, v in self._attrs.items() if k not in drop_keys and len(v) == 1}
        if callable(rename):
            self._attrs = rename(self._attrs)

    @property
    def attrs(self):
        return self._attrs


class TensorBackendArray(xarray.backends.common.BackendArray):
    def __init__(self, tensor, dims, shape, xp, dtype, var_name):
        super().__init__()
        self.tensor = tensor
        self.dims = dims
        self.shape = shape
        self._var_name = var_name
        self.dtype = dtype
        self.xp = xp

        from dask.utils import SerializableLock

        self.lock = SerializableLock()

    @property
    def nbytes(self):
        from math import prod

        return prod(self.shape) * self.dtype.itemsize

    def __getitem__(self, key: xarray.core.indexing.ExplicitIndexer):
        # indexing_support = indexing.IndexingSupport.BASIC
        # raw_key, numpy_indices = indexing.decompose_indexer(
        #     key, self.shape, indexing_support
        # )
        # result = self._raw_indexing_method(raw_key.tuple)
        # if numpy_indices.tuple:
        #     # index the loaded np.ndarray
        #     result = indexing.NdArrayLikeIndexingAdapter(result)[numpy_indices]
        # return result
        return indexing.explicit_indexing_adapter(
            key,
            self.shape,
            indexing.IndexingSupport.BASIC,
            self._raw_indexing_method,
        )
        # bug in xarray here? tries to create a NumpyIndexingAdapter instead of NdArrayLikeIndexingAdapter
        # patched in local copy for now, but could construct this ourself

    def _raw_indexing_method(self, key: tuple):
        with self.lock:
            # LOG.debug(f"TensorBackendArray._raw_indexing_method var={self._var_name}")
            # LOG.debug(f"   dims={self.dims} key={key} shape={self.shape}")
            # LOG.debug(f"   tensor.user_coords={self.tensor.user_coords}")

            r = self.tensor[key]
            # LOG.debug(f"   cubelet user_shape={r.user_shape}")

            # LOG.debug(f"   {r.user_shape=}")
            field_index = r.field_indexes(key)
            if self.tensor.is_full_field(field_index):
                field_index = None

            # LOG.debug(f"   {field_index=}")

            try:
                result = r.to_array(index=field_index, array_backend=self.xp, dtype=self.dtype)
            except Exception as e:
                LOG.exception("Error in to_array:", e)
                raise

            # LOG.debug(f"   {result.shape=}"

            # ensure axes are squeezed when needed
            singles = [i for i in list(range(len(r.user_shape))) if isinstance(key[i], int)]
            if singles:
                result = result.squeeze(axis=tuple(singles))

            return result


class BackendDataBuilder(metaclass=ABCMeta):
    def __init__(self, ds, profile, dims, grid=None, fixed_local_attrs=None):
        self.ds = ds
        self.profile = profile
        self.dims = dims

        self.flatten_values = profile.flatten_values

        # Array backend/namespace
        array_backend = profile.array_backend
        if array_backend is None:
            array_backend = "numpy"

        from earthkit.utils.array import get_backend

        self.array_backend = get_backend(array_backend)
        assert self.array_backend is not None, f"Unsupported array_backend : {array_backend}"

        dtype = profile.dtype
        if dtype is None:
            dtype = "float64"
        self.dtype = self.array_backend.make_dtype(dtype)

        # Note: these coords inside the tensor are called user_coords and
        # the corresponding dims are called user_dims
        self.tensor_coords = {}

        # coords describing the field dimensions
        self.field_coords = {}

        self.fixed_local_attrs = ensure_dict(fixed_local_attrs)

        self.grid = self._ensure_grid(grid)
        if self.profile.add_geo_coords:
            self.field_coords = self._make_field_coords()

    def coords(self):
        r = {k: v.to_xr_var(self.profile) for k, v in self.tensor_coords.items()}
        r.update(self.field_coords)
        return r

    def _ensure_grid(self, grid):
        if grid is None:
            from .grid import TensorGrid

            grid = TensorGrid(self.ds[0], self.flatten_values)
        return grid

    def _make_field_coords(self):
        r = {}
        for k, v in self.grid.coords.items():
            dims = {x: self.grid.dims[x] for x in self.grid.coords_dim[k]}
            r[k] = xarray.Variable(dims, v, self.profile.attrs.coord_attrs.get(k, {}))
        return r

    def collect_date_coords(self, tensor):
        if (
            self.profile.add_valid_time_coord
            and "valid_time" not in tensor.user_dims
            and "valid_datetime" not in tensor.user_coords
            and "valid_time" not in self.tensor_coords
        ):
            from .coord import Coord

            _dims, _vals = tensor.make_valid_datetime(self.dims)
            if _dims is not None and _vals is not None:
                self.tensor_coords["valid_time"] = Coord.make("valid_time", _vals, dims=_dims)

    def build(self):
        # we assume each variable forms a full cube
        var_builders = self.pre_build_variables()

        # build variable and global attributes
        xr_attrs = self.profile.attrs.builder.build(self.ds, var_builders, rename=True)
        xr_coords = self.coords()

        # build variables
        xr_vars = {
            self.profile.variable.rename(k): v.build(add_earthkit_attrs=self.profile.add_earthkit_attrs)
            for k, v in var_builders.items()
        }

        # build dataset
        dataset = xarray.Dataset(xr_vars, coords=xr_coords, attrs=xr_attrs)

        dataset = self.profile.rename_dataset_dims(dataset)

        # dim_map = self.profile.rename_dims_map()
        # if dim_map:
        #     d = {k: v for k, v in dim_map.items() if k in dataset.dims}
        #     if d:
        #         dataset = dataset.rename(d)

        if "source" not in dataset.encoding:
            dataset.encoding["source"] = None

        return dataset

    @abstractmethod
    def build_values(self, *args, **kwargs):
        pass

    @abstractmethod
    def pre_build_variables(self):
        pass

    def pre_build_variable(self, ds_var, dims, name):
        tensor_dims, tensor_coords, tensor_coords_component, tensor_extra_attrs = self.prepare_tensor(
            ds_var, dims, name
        )

        tensor_dim_keys = [d.key for d in tensor_dims]

        # LOG.debug(f"{tensor_dims=} {tensor_coords=} {tensor_coords_component=} {tensor_extra_attrs=}")
        # LOG.debug(f"{tensor_dim_keys=}")

        tensor = ds_var.to_tensor(
            *tensor_dim_keys,
            sort=False,
            progress_bar=False,
            user_dims_and_coords=tensor_coords,
            field_dims_and_coords=(self.grid.dims, self.grid.coords),
            flatten_values=self.flatten_values,
        )

        var_dims = []
        for d in tensor_dims:
            # Create coord for each dimension
            # Dimensions like "level_per_type" are templates and will be
            # added as multiple concrete dimensions to the dataset
            k, c = d.as_coord(
                d.key, tensor.user_coords[d.key], tensor_coords_component.get(d.key, None), tensor.source
            )
            if k not in self.tensor_coords:
                self.tensor_coords[k] = c
            var_dims.append(k)
        var_dims.extend(tensor.field_dims)

        self.collect_date_coords(tensor)
        data_maker = self.build_values
        remapping = self.profile.remapping.build()

        var_builder = VariableBuilder(
            name,
            var_dims,
            data_maker,
            tensor,
            remapping,
            local_attr_keys=tensor_extra_attrs,
            fixed_local_attrs=self.fixed_local_attrs,
        )

        return var_builder

    def prepare_tensor(self, ds, dims, name):
        tensor_dims = []
        tensor_coords = {}
        tensor_coords_component = {}
        tensor_extra_attrs = []

        # LOG.debug(f"{name=} {dims=}")

        vals, component_vals = ds.unique_values(
            [d.key for d in dims], component=self.profile.add_earthkit_attrs
        )

        # LOG.debug(f"unique_values={vals}")
        # LOG.debug(f"ensure_dims={self.profile.dims.ensure_dims}")
        # LOG.debug(f"dims_as_attrs={self.profile.dims.dims_as_attrs}")

        for d in dims:
            num = len(vals[d.key])
            # LOG.debug(f"  {d.key=} {vals[d.key]}")
            if num == 0:
                continue
                # if d.name not in self.profile.dims.ensure_dims:
                #     raise ValueError(f"Dimension {d} has no valid values for variable={name}")
            else:
                if num > 1 and d.enforce_unique:
                    raise ValueError(
                        f"Dimension '{d.name}' of variable '{name}' cannot have multiple values={vals[d.key]}"
                    )
                elif num == 1 and d.name in self.profile.dims.dims_as_attrs:
                    tensor_extra_attrs.append(d.key)
                elif num > 1 or not self.profile.dims.squeeze or d.name in self.profile.dims.ensure_dims:
                    tensor_dims.append(d)
                    tensor_coords[d.key] = vals[d.key]
                    if component_vals and d.key in component_vals:
                        tensor_coords_component[d.key] = component_vals[d.key]

                    # check if the dims/coords are consistent with the tensors of
                    # the previous variables
                    self.check_tensor_coords(name, d.key, tensor_coords)

        # TODO:  check if fieldlist forms a full hypercube with respect to the the dims/coordinates
        return tensor_dims, tensor_coords, tensor_coords_component, tensor_extra_attrs

    def check_tensor_coords(self, var_name, coord_name, tensor_coords):
        from .check import check_coords

        check_coords(var_name, coord_name, tensor_coords, self.tensor_coords)


class TensorBackendDataBuilder(BackendDataBuilder):
    """Build a dataset using memory backend.

    Each variable contain a TensorBackendArray object that loads the data on demand from
    the GRIB using the associated TensorFieldList. This solution can work bot for GRIB data
    stored in memory or on disk.
    """

    def pre_build_variables(self):
        """Generate a builder for each variable"""
        builders = {}

        if self.profile.variable.is_mono:
            name = self.profile.variable.name
            builders[name] = self.pre_build_variable(self.ds, self.dims, name)
        else:
            # we assume each variable forms a full cube
            key = self.profile.variable.key
            for name in self.profile.variable.variables:
                ds_var = self.ds.sel(**{key: name})
                builders[name] = self.pre_build_variable(ds_var, self.dims, name)

        return builders

    def build_values(self, tensor, var_dims, name):
        """Generate the data object stored in the xarray variable"""
        # There is no need for the extra structures in the wrapped source in the
        # tensor any longer. It is replaced by the original unwrapped fieldlist.
        tensor.source = tensor.source.unwrap()

        backend_array = TensorBackendArray(
            tensor,
            var_dims,
            tensor.full_shape,
            self.array_backend.namespace,
            self.dtype,
            name,
        )

        data = indexing.LazilyIndexedArray(backend_array)
        return data


class MemoryBackendDataBuilder(BackendDataBuilder):
    """Build a dataset using memory backend.

    Each variable contain data values in memory copied from the input fields. As soon as a
    field's values are copied, the field is immediately released and its memory is freed if the
    field supports this operation. Fields released in this way are not available for further access.
    """

    def pre_build_variables(self):
        """Generate a builder for each variable"""
        builders = {}

        if self.profile.variable.is_mono:
            name = self.profile.variable.name
            builders[name] = self.pre_build_variable(self.ds, self.dims, name)
        else:
            groups = self.ds.group(self.profile.variable.key, self.profile.variable.variables)

            # we assume each variable forms a full cube
            for name in groups:
                ds_var = groups[name]
                builders[name] = self.pre_build_variable(ds_var, self.dims, name)

        return builders

    def build_values(self, tensor, var_dims, name):
        """Generate the data object stored in the xarray variable"""

        # At this point all the fields must be a ReleasableField.
        # We mark the fields so that their data will be released on the next
        # values access.
        if self.profile.release_source:
            for f in tensor.source:
                f.keep = False

        return tensor.to_array(dtype=self.dtype, array_backend=self.array_backend)


class DatasetBuilder:
    def __init__(
        self,
        ds,
        profile,
        backend_kwargs=None,
    ):
        self.ds = ds
        backend_kwargs = ensure_dict(backend_kwargs)
        assert "profile" not in backend_kwargs

        if isinstance(profile, Profile):
            self.profile = profile
        else:
            self.profile = Profile.make(profile, **backend_kwargs)

        backend_kwargs = ensure_dict(backend_kwargs)
        self.profile_kwargs = dict(**backend_kwargs)

        # LOG.debug(f"{self.profile.name=}")
        # LOG.debug(f"{backend_kwargs=}")
        if self.profile.lazy_load:
            self.builder = TensorBackendDataBuilder
        else:
            self.builder = MemoryBackendDataBuilder

        self.split_dims = self.profile.dims.split_dims
        self.direct_backend = self.profile.direct_backend

        self.grids = {}

    def parse(self, ds, profile=None, full=False):
        assert not hasattr(ds, "_ek_builder")
        from .fieldlist import XArrayInputFieldList

        if profile is None:
            profile = self.profile.copy()

        remapping = profile.remapping.build()

        # LOG.debug(f"{remapping=}")
        # LOG.debug(f"{profile.remapping=}")
        LOG.debug(f"{profile.index_keys=}")

        # create a new fieldlist for optimised access to unique values
        ds_xr = XArrayInputFieldList(
            ds, keys=profile.index_keys, remapping=remapping, component=profile.add_earthkit_attrs
        )
        # LOG.debug(f"{ds.db=}")

        # LOG.debug(f"before update: {profile.dim_keys=}")
        profile.update(ds_xr)
        # LOG.debug(f"after update: {profile.dim_keys=}")

        # LOG.debug(f"{profile.sort_keys=}")
        # the data is only sorted once
        ds_xr = ds_xr.order_by(profile.sort_keys)

        if not profile.lazy_load and profile.release_source:
            ds_xr.make_releasable()

        return ds_xr, profile

    def grid(self, ds):
        grids = ds.index("md5GridSection")

        if not grids:
            grid = "_custom_" + str(id(ds))
        else:
            # if len(grids) != 1:
            #     raise ValueError(f"Expected one grid, got {len(grids)}")
            grid = grids[0]

        key = (grid, self.profile.flatten_values)

        if key not in self.grids:
            from .grid import TensorGrid

            self.grids[key] = TensorGrid(ds[0], self.profile.flatten_values)
        return self.grids[key]


class SingleDatasetBuilder(DatasetBuilder):
    def __init__(self, *args, from_xr=False, **kwargs):
        super().__init__(*args, **kwargs)

        if self.split_dims:
            raise ValueError("SingleDatasetMaker does not support splitting")

        if from_xr and self.direct_backend:
            raise ValueError(
                "SingleDatasetMaker does not support direct_backend=True when invoked from xarray"
            )

    def build(self):
        ds_sorted, _ = self.parse(self.ds, self.profile)
        dims = self.profile.dims.to_list()
        LOG.debug(f"{dims=}")
        builder = self.builder(
            ds_sorted,
            self.profile,
            dims,
            grid=self.grid(ds_sorted),
        )
        r = builder.build()
        return r


class SplitDatasetBuilder(DatasetBuilder):
    def __init__(self, *args, backend_kwargs=None, other_kwargs=None):
        """
        split_dims: str, or iterable of str, None
            Dimension or list of dimensions to use for splitting the data into multiple hypercubes.
            Default is None.
        """
        super().__init__(*args, backend_kwargs=backend_kwargs)

        if not self.direct_backend:
            self.xr_open_dataset_kwargs = ensure_dict(other_kwargs)
            backend_kwargs = {}
            backend_kwargs["profile"] = self.profile
            self.xr_open_dataset_kwargs["backend_kwargs"] = backend_kwargs

        if not self.split_dims:
            raise ValueError("SplitDatasetMaker requires split_dims")

    def prepare(self, keys):
        from .fieldlist import XArrayInputFieldList

        remapping = self.profile.remapping.build()

        # LOG.debug(f"split_dims={self.split_dims}")
        ds_xr = XArrayInputFieldList(self.ds, keys=self.profile.index_keys, remapping=remapping)

        vals, _ = ds_xr.unique_values(keys)
        LOG.debug(f"{keys=}, {vals=}")

        return ds_xr, vals

    def build(self):
        from .splitter import Splitter

        splitter = Splitter.make(self.split_dims)
        datasets = []
        split_coords_list = []
        for ds, profile, split_coords in splitter.split(self):
            dims = profile.dims.to_list()
            split_coords_list.append(dict(split_coords))
            LOG.debug(f"splitting {dims=} type of s_ds={type(ds)} {split_coords=}")
            split_coords.pop(profile.variable.key, None)
            builder = self.builder(ds, profile, dims, grid=self.grid(ds), fixed_local_attrs=split_coords)

            ds._ek_builder = builder
            if self.direct_backend:
                datasets.append(builder.build())
            else:
                ds._ek_builder = builder
                datasets.append(xarray.open_dataset(ds, **self.xr_open_dataset_kwargs))
                ds._ek_builder = None

        return datasets, split_coords_list


def from_earthkit(ds, backend_kwargs=None, other_kwargs=None):
    """Create an xarray dataset from an earthkit fieldlist.

    Parameters
    ----------
    ds: FieldList
        The input fieldlist.
    backend_kwargs: dict, optional
        Backend kwargs that can be passed to
        :py:meth:`xarray.open_dataset` as "backend_kwargs".
    other_kwargs: dict, optional
        Additional kwargs passed to :py:meth:`xarray.open_dataset`. Cannot contain
        any of the keys in ``backend_kwargs``.
    """
    backend_kwargs = ensure_dict(backend_kwargs)
    other_kwargs = ensure_dict(other_kwargs)

    # certain kwargs are both backend_kwargs and other_kwargs. We need all of
    # these for the profile
    profile_kwargs = dict(**backend_kwargs)
    for k in BACKEND_AND_XR_OPEN_DS_KWARGS:
        if k in other_kwargs:
            profile_kwargs[k] = other_kwargs[k]

    # to create the profile we need all the possible backend_kwargs (bar profile)
    profile = profile_kwargs.pop("profile", Profile.DEFAULT_PROFILE_NAME)
    profile = Profile.make(profile, **profile_kwargs)

    # the backend builder is directly called bypassing xarray.open_dataset
    if profile.direct_backend:
        if not profile.dims.split_dims:
            return SingleDatasetBuilder(ds, profile).build()
        else:
            return SplitDatasetBuilder(ds, profile).build()
    # xarray.open_dataset is called
    else:
        # LOG.debug(f"from_earthkit {backend_kwargs=} {profile_kwargs=}")
        assert other_kwargs["engine"] == "earthkit"
        if not profile.dims.split_dims:
            backend_kwargs["profile"] = profile
            # certain kwargs are not allowed in xarray.open_dataset
            for k in NON_XR_OPEN_DS_KWARGS:
                backend_kwargs.pop(k, None)
            return xarray.open_dataset(ds, backend_kwargs=backend_kwargs, **other_kwargs)
        else:
            return SplitDatasetBuilder(ds, profile, other_kwargs=other_kwargs).build()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import pytest

from earthkit.data.utils.bbox import BoundingBox


def test_globe():
    globe1 = BoundingBox(north=90, west=0, east=360, south=-90)
    assert globe1.width == 360
    assert globe1.west == 0

    globe2 = BoundingBox(north=90, west=-180, east=180, south=-90)
    assert globe2.width == 360
    assert globe2.west == -180

    assert BoundingBox.union([globe1, globe2]).width == 360, BoundingBox.union([globe1, globe2]).width
    assert BoundingBox.union([globe2, globe1]).width == 360, BoundingBox.union([globe2, globe1]).width

    b1 = BoundingBox(north=90, west=-180, east=0, south=-90)
    b2 = BoundingBox(north=90, west=0, east=180, south=-90)
    b0 = BoundingBox.union([b1, b2])
    assert b0.width == 360


def test_almost_globe():
    globe1 = BoundingBox(north=90, west=1, east=360, south=-90)
    globe2 = BoundingBox(north=90, west=-180, east=179, south=-90)

    assert BoundingBox.union([globe1, globe2]).width == 360, BoundingBox.union([globe1, globe2]).width
    assert BoundingBox.union([globe2, globe1]).width == 360, BoundingBox.union([globe2, globe1]).width


def test_bbox():
    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i, south=30, east=10 + i)
        assert bbox.width == 10, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i, south=30, east=350 + i)
        assert bbox.width == 350, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i, south=30, east=(10 + i) % 360)
        assert bbox.width == 10, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i, south=30, east=(350 + i) % 360)
        assert bbox.width == 350, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i % 360, south=30, east=10 + i)
        assert bbox.width == 10, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i % 360, south=30, east=350 + i)
        assert bbox.width == 350, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i % 360, south=30, east=(10 + i) % 360)
        assert bbox.width == 10, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i % 360, south=30, east=(350 + i) % 360)
        assert bbox.width == 350, bbox

    for i in range(-365, 365):
        bbox1 = BoundingBox(north=90, west=150 + i, south=30, east=170 + i)
        bbox2 = BoundingBox(north=90, west=-170 + i, south=30, east=-150 + i)

        bbox = BoundingBox.union([bbox1, bbox2])
        assert bbox.width == 60, (bbox1, bbox2, bbox, bbox.width)

        bbox = BoundingBox.union([bbox2, bbox1])
        assert bbox.width == 60, (bbox1, bbox2, bbox, bbox.width)

        merge1 = BoundingBox.union([bbox1, bbox2])
        merge2 = BoundingBox.union([bbox2, bbox1])
        assert merge1 == merge2, (bbox1, bbox2, merge1, merge2)

    with pytest.raises(ValueError):
        BoundingBox(north=-10, west=0, south=30, east=1)

    b0 = BoundingBox(north=89.9746, west=-179.975, south=-89.9746, east=179.975)
    b1 = BoundingBox(north=89.9746, west=-179.975, south=-89.9746, east=179.975)
    b2 = BoundingBox(north=89.9746, west=-179.975, south=-89.9746, east=179.975)
    b3 = BoundingBox.union([b1, b2])

    assert b0 == b3


def test_overlapping_bbox_1():
    for offset in range(-500, 500, 10):
        one = BoundingBox(north=90, west=offset + 10, east=offset + 20, south=-90)
        two = BoundingBox(north=90, west=offset + 40, east=offset + 60, south=-90)
        three = BoundingBox(north=90, west=offset + 15, east=offset + 35, south=-90)

        sets = []
        sets.append([one, two, three])
        sets.append([two, one, three])
        sets.append([one, three, two])
        sets.append([one, three, two, one, three])
        sets.append([one, one, one, two, one, two, one, three])
        for i, s in enumerate(sets):
            merged = BoundingBox.union(s)
            expected = BoundingBox(east=offset + 60, west=offset + 10, north=90, south=-90)
            assert merged.east == expected.east, (
                i,
                merged.east,
                expected.east,
                merged,
                s,
            )
            assert merged.west == expected.west, (
                i,
                merged.west,
                expected.west,
                merged,
                s,
            )

        four = BoundingBox(north=90, west=offset - 200, east=offset + 10, south=-90)
        sets = []
        sets.append([one, two, three, four])
        sets.append([two, one, four, three])
        sets.append([one, three, four, two])
        sets.append([one, three, two, four, one, three])
        sets.append([one, one, one, two, four, one, two, one, three])
        for i, s in enumerate(sets):
            merged = BoundingBox.union(s)
            expected = BoundingBox(east=offset + 60, west=offset - 200, north=90, south=-90)
            assert merged.east % 360 == expected.east % 360, (
                i,
                offset,
                merged.east,
                expected.east,
                merged,
                s,
            )
            assert merged.west % 360 == expected.west % 360, (
                i,
                merged.west,
                expected.west,
                merged,
                s,
            )
            assert merged.west < merged.east


def xxxtest_overlaps():
    b1 = BoundingBox(north=90, west=-200, south=-90, east=-130)
    b2 = BoundingBox(north=90, west=-180, south=-90, east=-90)
    b0 = b1.overlaps(b2)
    assert b0


def test_overlapping_bbox_2():
    b1 = BoundingBox(north=90, west=-200, south=-90, east=-130)
    b2 = BoundingBox(north=90, west=-180, south=-90, east=-90)
    b0 = BoundingBox.union([b1, b2])

    assert b0.width == 110, b0.width

    b3 = BoundingBox(north=90, west=-210, south=-90, east=-160)
    b0 = BoundingBox.union([b0, b3])
    assert b0.width == 120, b0.width

    #      ----------------------
    #           ---------------------
    # ---------------------

    for i in range(-365, 365):
        b1 = BoundingBox(north=90, west=10 + i, south=-90, east=80 + i)
        b2 = BoundingBox(north=90, west=30 + i, south=-90, east=120 + i)
        b3 = BoundingBox(north=90, west=0 + i, south=-90, east=50 + i)
        b0 = BoundingBox.union([b1, b2, b3])
        assert b0.width == 120, (b0.width, b0)

    # --------------------------------
    #           ---------------------
    #     ---------------------

    for i in range(-365, 365):
        b1 = BoundingBox(north=90, west=-10 + i, south=-90, east=200 + i)
        b2 = BoundingBox(north=90, west=30 + i, south=-90, east=120 + i)
        b3 = BoundingBox(north=90, west=0 + i, south=-90, east=50 + i)

        b0 = BoundingBox.union([b1, b2, b3])
        assert b0.width == 210, (b0.width, b0)

    # --------------------------------
    #           --------------------------
    #     ---------------------

    for i in range(-365, 365):
        b1 = BoundingBox(north=90, west=-10 + i, south=-90, east=200 + i)
        b2 = BoundingBox(north=90, west=30 + i, south=-90, east=300 + i)
        b3 = BoundingBox(north=90, west=0 + i, south=-90, east=50 + i)

        b0 = BoundingBox.union([b1, b2, b3])

        assert b0.width == 310, (b0.width, b0)


def test_overlapping_bbox_3():
    for i in range(361):
        b1 = BoundingBox(north=90, west=-45 - i, south=-90, east=45 - i)
        b2 = BoundingBox(north=90, west=-45 + i, south=-90, east=45 + i)
        b0 = BoundingBox.union([b1, b2])

        if i <= 90:
            assert b0.width == 90 + 2 * i, (i, b0.width, b0)

        if i > 90 and i <= 180:
            assert b0.width == 3 * 90 - 2 * (i - 90), (i, b0.width, b0)

        if i > 180 and i <= 270:
            assert b0.width == 90 + 2 * (i - 180), (i, b0.width, b0)

        if i > 270:
            assert b0.width == 3 * 90 - 2 * (i - 270), (i, b0.width, b0)


def test_bbox_union_with():
    b1 = BoundingBox(north=50, west=-20, south=-30, east=40)
    b2 = BoundingBox(north=60, west=-10, south=-40, east=55)
    b = b1.union_with(b2)
    assert b == BoundingBox(north=60, west=-20, south=-40, east=55)


def test_bbox_make_invalid():
    import numpy as np

    b = BoundingBox.make_invalid()
    assert b.as_tuple() == (np.nan, np.nan, np.nan, np.nan)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.core.fieldlist import FieldList
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import write_to_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from array_fl_fixtures import check_array_fl  # noqa: E402
from array_fl_fixtures import check_array_fl_from_to_fieldlist  # noqa: E402


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_array_fl_grib_single_field(write_method):
    ds = from_source("file", earthkit_examples_file("test.grib"))

    assert ds[0].metadata("shortName") == "2t"

    lat, lon, v = ds[0].data(flatten=True)
    v1 = v + 1

    md = ds[0].metadata()
    md1 = md.override(shortName="msl")
    r = FieldList.from_numpy(v1, md1)

    def _check_field(r):
        assert len(r) == 1
        assert np.allclose(r[0].values, v1)
        assert r[0].shape == ds[0].shape
        assert r[0].metadata("shortName") == "msl"
        _lat, _lon, _v = r[0].data(flatten=True)
        assert np.allclose(_lat, lat)
        assert np.allclose(_lon, lon)
        assert np.allclose(_v, v1)

    _check_field(r)

    # save to disk
    tmp = temp_file()
    write_to_file(write_method, tmp.path, r)
    assert os.path.exists(tmp.path)
    r_tmp = from_source("file", tmp.path)
    _check_field(r_tmp)


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_array_fl_grib_multi_field(write_method):
    ds = from_source("file", earthkit_examples_file("test.grib"))

    assert ds[0].metadata("shortName") == "2t"

    v = ds.values
    v1 = v + 1

    md1 = [f.metadata().override(shortName="2d") for f in ds]
    r = FieldList.from_numpy(v1, md1)

    assert len(r) == 2
    assert np.allclose(v1, r.values)
    for i, f in enumerate(r):
        assert f.shape == ds[i].shape
        assert f.metadata("shortName") == "2d", f"shortName {i}"
        assert f.metadata("name") == "2 metre dewpoint temperature", f"name {i}"

    # save to disk
    tmp = temp_file()
    write_to_file(write_method, tmp.path, r)
    assert os.path.exists(tmp.path)
    r_tmp = from_source("file", tmp.path)
    assert len(r_tmp) == 2
    assert np.allclose(v1, r_tmp.values)
    for i, f in enumerate(r_tmp):
        assert f.shape == ds[i].shape
        assert f.metadata("shortName") == "2d", f"shortName {i}"
        assert f.metadata("name") == "2 metre dewpoint temperature", f"name {i}"


def test_array_fl_grib_from_list_of_arrays():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    md_full = ds.metadata("param")
    assert len(ds) == 2

    v = [ds[0].values, ds[1].values]
    md = [f.metadata().override(generatingProcessIdentifier=150) for f in ds]
    r = FieldList.from_numpy(v, md)

    check_array_fl(r, [ds], md_full)


def test_array_fl_grib_from_list_of_arrays_bad():
    ds = from_source("file", earthkit_examples_file("test.grib"))

    v = ds[0].values
    md = [f.metadata().override(generatingProcessIdentifier=150) for f in ds]

    with pytest.raises(ValueError):
        _ = FieldList.from_numpy(v, md)

    with pytest.raises(ValueError):
        _ = FieldList.from_numpy([v], md)


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"dtype": np.float32},
        {"flatten": False},
        {"flatten": True},
        {"flatten": True, "dtype": np.float32},
    ],
)
def test_array_fl_grib_from_to_fieldlist(kwargs):
    ds = from_source("file", earthkit_examples_file("test.grib"))
    md_full = ds.metadata("param")
    assert len(ds) == 2

    r = ds.to_fieldlist(array_backend="numpy", **kwargs)
    check_array_fl_from_to_fieldlist(r, [ds], md_full, **kwargs)


def test_array_fl_grib_from_to_fieldlist_repeat():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    md_full = ds.metadata("param")
    assert len(ds) == 2

    kwargs = {}
    r = ds.to_fieldlist(array_backend="numpy", **kwargs)
    check_array_fl_from_to_fieldlist(r, [ds], md_full, **kwargs)

    kwargs = {"flatten": True, "dtype": np.float32}
    r1 = r.to_fieldlist(array_backend="numpy", **kwargs)
    assert r1 is not r
    check_array_fl_from_to_fieldlist(r1, [ds], md_full, **kwargs)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import os
import sys

import numpy as np
import pytest

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from forcings_fixtures import load_forcings_fs  # noqa: E402


def check_array(v, shape=None, first=None, last=None, meanv=None, eps=1e-3):
    assert v.shape == shape
    assert np.isclose(v[0], first, eps)
    assert np.isclose(v[-1], last, eps)
    assert np.isclose(v.mean(), meanv, eps)


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
def test_forcings_values(input_data):
    ds, _ = load_forcings_fs(last_step=12, input_data=input_data)
    eps = 1e-5
    num = len(ds)

    # whole file
    v = ds.values
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (num, 209)
    vf = v[0].flatten()
    check_array(
        vf,
        (209,),
        first=0.95630476,
        last=0.54463904,
        meanv=0.7793134730178092,
        eps=eps,
    )

    vf = v[num - 1].flatten()
    check_array(
        vf,
        (209,),
        first=0.89100652,
        last=0.70710678,
        meanv=0.917041819159978,
        eps=eps,
    )


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
def test_forcings_to_numpy(input_data):
    ds, _ = load_forcings_fs(last_step=12, input_data=input_data)
    eps = 1e-5
    num = len(ds)

    # whole file
    v = ds.to_numpy(flatten=True)
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (num, 209)
    vf = v[0].flatten()
    check_array(
        vf,
        (209,),
        first=0.95630476,
        last=0.54463904,
        meanv=0.7793134730178092,
        eps=eps,
    )

    vf = v[num - 1].flatten()
    check_array(
        vf,
        (209,),
        first=0.89100652,
        last=0.70710678,
        meanv=0.917041819159978,
        eps=eps,
    )


@pytest.mark.parametrize(
    "options, expected_shape",
    [
        (
            {},
            (
                16,
                11,
                19,
            ),
        ),
        (
            {"flatten": True},
            (
                16,
                209,
            ),
        ),
        ({"flatten": False}, (16, 11, 19)),
    ],
)
def test_forcings_to_numpy_shape(options, expected_shape):
    ds, _ = load_forcings_fs(last_step=12)
    num = 16
    eps = 1e-5

    # whole file
    v = ds.to_numpy()
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (num, 11, 19)
    vf0 = ds[0].to_numpy().flatten()
    assert vf0.shape == (209,)
    vfm1 = ds[num - 1].to_numpy().flatten()
    assert vfm1.shape == (209,)

    v1 = ds.to_numpy(**options)
    assert isinstance(v1, np.ndarray)
    assert v1.dtype == np.float64
    assert v1.shape == expected_shape
    vr = v1[0].flatten()
    assert np.allclose(vf0, vr, eps)
    vr = v1[num - 1].flatten()
    assert np.allclose(vfm1, vr, eps)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_forcings_to_numpy_dtype(dtype):
    ds, _ = load_forcings_fs(last_step=12)

    v = ds[0].to_numpy(dtype=dtype)
    assert v.dtype == dtype

    v = ds.to_numpy(dtype=dtype)
    assert v.dtype == dtype


@pytest.mark.parametrize(
    "kwarg,expected_shape,expected_dtype",
    [
        ({}, (11, 19), np.float64),
        ({"flatten": True}, (209,), np.float64),
        ({"flatten": True, "dtype": np.float32}, (209,), np.float32),
        ({"flatten": True, "dtype": np.float64}, (209,), np.float64),
        ({"flatten": False}, (11, 19), np.float64),
        ({"flatten": False, "dtype": np.float32}, (11, 19), np.float32),
        ({"flatten": False, "dtype": np.float64}, (11, 19), np.float64),
    ],
)
def test_forcings_field_data(kwarg, expected_shape, expected_dtype):
    ds, _ = load_forcings_fs(params=["longitude"], last_step=12)

    latlon = ds[0].to_latlon(**kwarg)
    v = ds[0].to_numpy(**kwarg)

    d = ds[0].data(**kwarg)
    assert isinstance(d, np.ndarray)
    assert d.dtype == expected_dtype
    assert len(d) == 3
    assert d[0].shape == expected_shape
    assert np.allclose(d[0], latlon["lat"])
    assert np.allclose(d[1], latlon["lon"])
    assert np.allclose(d[2], v)

    d = ds[0].data(keys="lat", **kwarg)
    assert d.shape == expected_shape
    assert d.dtype == expected_dtype
    assert np.allclose(d, latlon["lat"])

    d = ds[0].data(keys="lon", **kwarg)
    assert d.shape == expected_shape
    assert d.dtype == expected_dtype
    assert np.allclose(d, latlon["lon"])

    d = ds[0].data(keys="value", **kwarg)
    assert d.shape == expected_shape
    assert d.dtype == expected_dtype
    assert np.allclose(d, v)

    d = ds[0].data(keys=("value", "lon"), **kwarg)
    assert isinstance(d, np.ndarray)
    assert d.dtype == expected_dtype
    assert len(d) == 2
    assert np.allclose(d[0], v)
    assert np.allclose(d[1], latlon["lon"])


@pytest.mark.parametrize(
    "kwarg,expected_shape,expected_dtype",
    [
        ({}, (11, 19), np.float64),
        ({"flatten": True}, (209,), np.float64),
        ({"flatten": True, "dtype": np.float32}, (209,), np.float32),
        ({"flatten": True, "dtype": np.float64}, (209,), np.float64),
        ({"flatten": False}, (11, 19), np.float64),
        ({"flatten": False, "dtype": np.float32}, (11, 19), np.float32),
        ({"flatten": False, "dtype": np.float64}, (11, 19), np.float64),
    ],
)
def test_forcings_fieldlist_data(kwarg, expected_shape, expected_dtype):
    ds, _ = load_forcings_fs(params=["longitude"], last_step=12)
    num = 2

    latlon = ds.to_latlon(**kwarg)
    v = ds.to_numpy(**kwarg)

    d = ds.data(**kwarg)
    assert isinstance(d, np.ndarray)
    assert d.shape == tuple([num + 2, *expected_shape])
    assert d.dtype == expected_dtype
    assert np.allclose(d[0], latlon["lat"])
    assert np.allclose(d[1], latlon["lon"])
    assert np.allclose(d[2], v[0])
    assert np.allclose(d[3], v[1])

    d = ds.data(keys="lat", **kwarg)
    assert d.shape == tuple([1, *expected_shape])
    assert d.dtype == expected_dtype
    assert np.allclose(d[0], latlon["lat"])

    d = ds.data(keys="lon", **kwarg)
    assert d.shape == tuple([1, *expected_shape])
    assert d.dtype == expected_dtype
    assert np.allclose(d[0], latlon["lon"])

    d = ds.data(keys="value", **kwarg)
    assert d.shape == tuple([num, *expected_shape])
    assert d.dtype == expected_dtype
    assert np.allclose(d, v)

    d = ds.data(keys=("value", "lon"), **kwarg)
    assert isinstance(d, np.ndarray)
    assert d.shape == tuple([num + 1, *expected_shape])
    assert d.dtype == expected_dtype
    assert np.allclose(d[0], v[0])
    assert np.allclose(d[1], v[1])
    assert np.allclose(d[2], latlon["lon"])


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest
import yaml

from earthkit.data.testing import earthkit_test_data_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from forcings_fixtures import all_params  # noqa: E402
from forcings_fixtures import load_forcings_fs  # noqa: E402


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
def _build_proc_ref(input_data):
    import yaml

    ds, _ = load_forcings_fs(params=all_params, last_step=12, input_data=input_data)
    d = {}
    for p in all_params:
        # print(f"p={p}")
        f = ds.sel(param=p, valid_datetime="2020-05-13T18:00:00")
        v = f[0].values
        d[p] = {
            "first": float(v[0]),
            "last": float(v[-1]),
            "mean": float(np.nanmean(v)),
        }

    # print(d)
    with open("_dev/proc.yaml", "w") as outfile:
        yaml.dump(d, outfile, sort_keys=True)


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
def test_forcings_proc(input_data):
    with open(earthkit_test_data_file(os.path.join("forcings", "proc.yaml")), "r") as f:
        ref = yaml.safe_load(f)

    ds, _ = load_forcings_fs(params=all_params, last_step=12, input_data=input_data)

    for p in all_params:
        f = ds.sel(param=p, valid_datetime="2020-05-13T18:00:00")
        assert len(f) == 1
        v = f[0].values
        r = ref[p]
        assert np.isclose(v[0], r["first"])
        assert np.isclose(v[-1], r["last"])
        assert np.isclose(np.nanmean(v), r["mean"])


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
@pytest.mark.parametrize("param,coord", [("latitude", "lat"), ("longitude", "lon")])
def test_forcings_proc_latlon(input_data, param, coord):
    ds, _ = load_forcings_fs(params=all_params, last_step=12, input_data=input_data)

    latlon = ds[0].to_latlon(flatten=True)
    coord = latlon[coord]
    date = "2020-05-13T18:00:00"

    f = ds.sel(param=param, valid_datetime=date)
    assert len(f) == 1
    assert np.allclose(f[0].values, coord)

    f = ds.sel(param="cos_" + param, valid_datetime=date)
    assert len(f) == 1
    assert np.allclose(f[0].values, np.cos(np.deg2rad(coord)))

    f = ds.sel(param="sin_" + param, valid_datetime=date)
    assert len(f) == 1
    assert np.allclose(f[0].values, np.sin(np.deg2rad(coord)))


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import pickle
import sys

import numpy as np
import pytest

from earthkit.data import config
from earthkit.data import from_source
from earthkit.data.core.temporary import temp_file
from earthkit.data.readers.grib.metadata import StandAloneGribMetadata
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import write_to_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from grib_fixtures import FL_FILE  # noqa: E402
from grib_fixtures import FL_NUMPY  # noqa: E402
from grib_fixtures import load_grib_data  # noqa: E402


def _pickle(data, representation):
    if representation == "file":
        with temp_file() as tmp:
            with open(tmp, "wb") as f:
                pickle.dump(data, f)

            with open(tmp, "rb") as f:
                data_res = pickle.load(f)
    elif representation == "memory":
        pickled_data = pickle.dumps(data)
        data_res = pickle.loads(pickled_data)
    else:
        raise ValueError(f"Invalid representation: {representation}")
    return data_res


@pytest.mark.parametrize("fl_type", FL_NUMPY)
@pytest.mark.parametrize("representation", ["file", "memory"])
def test_grib_serialise_metadata(fl_type, representation):
    ds, _ = load_grib_data("test.grib", fl_type)
    md = ds[0].metadata().override()

    md2 = _pickle(md, representation)

    keys = ["param", "date", "time", "step", "level", "gridType", "type"]
    for k in keys:
        assert md[k] == md2[k]


@pytest.mark.parametrize("representation", ["file", "memory"])
def test_grib_serialise_standalone_metadata(representation):
    ds = from_source("file", earthkit_examples_file("test.grib"))

    md_ref = {
        "param": "2t",
        "date": 20200513,
        "time": 1200,
        "step": 0,
        "level": 0,
        "gridType": "regular_ll",
        "type": "an",
    }

    md = StandAloneGribMetadata(ds[0].handle)
    for k, v in md_ref.items():
        assert md[k] == v

    md2 = _pickle(md, representation)
    for k, v in md_ref.items():
        assert md2[k] == v


@pytest.mark.parametrize("fl_type", FL_NUMPY)
@pytest.mark.parametrize("representation", ["file", "memory"])
def test_grib_serialise_array_field_memory(fl_type, representation):
    ds0, _ = load_grib_data("test.grib", fl_type)
    ds = ds0.to_fieldlist()

    for idx in range(len(ds)):
        f2 = _pickle(ds[idx], representation)

        assert np.allclose(ds[idx].values, f2.values), f"index={idx}"
        assert np.allclose(ds[idx].to_numpy(), f2.to_numpy()), f"index={idx}"

        keys = ["param", "date", "time", "step", "level", "gridType", "type"]
        for k in keys:
            assert ds[idx].metadata(k) == f2.metadata(k), f"index={idx}"


@pytest.mark.parametrize("fl_type", FL_NUMPY)
@pytest.mark.parametrize("representation", ["file", "memory"])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_serialise_array_fieldlist(fl_type, representation, write_method):
    ds0, _ = load_grib_data("test.grib", fl_type)
    ds = ds0.to_fieldlist()

    ds2 = _pickle(ds, representation)

    assert len(ds) == len(ds2)
    assert np.allclose(ds.values, ds2.values)

    keys = ["param", "date", "time", "step", "level", "gridType", "type"]
    for k in keys:
        ds2.metadata(k) == ds.metadata(k)

    r = ds2.sel(param="2t")
    assert len(r) == 1

    ds2[0]._array_ += 1
    v1 = ds[0]._array

    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds2)
        assert os.path.exists(tmp)
        r_tmp = from_source("file", tmp)
        assert len(ds2) == len(r_tmp)
        v_tmp = r_tmp[0].to_numpy()
        assert np.allclose(v1 + 1, v_tmp)


@pytest.mark.parametrize("fl_type", ["file"])
def test_grib_serialise_file_fieldlist_core(fl_type):
    ds, _ = load_grib_data("test.grib", fl_type)

    pickled_f = pickle.dumps(ds)
    ds2 = pickle.loads(pickled_f)

    assert len(ds) == len(ds2)
    assert np.allclose(ds.values, ds2.values)

    keys = ["param", "date", "time", "step", "level", "gridType", "type"]
    for k in keys:
        ds2.metadata(k) == ds.metadata(k)

    r = ds2.sel(param="2t")
    assert len(r) == 1


@pytest.mark.parametrize("fl_type", ["file"])
def test_grib_serialise_file_fieldlist_sel(fl_type):
    ds0, _ = load_grib_data("test6.grib", fl_type)
    ds = ds0.sel(param="t")
    assert len(ds) == 2

    pickled_f = pickle.dumps(ds)
    ds2 = pickle.loads(pickled_f)

    assert len(ds2) == 2
    assert np.allclose(ds.values, ds2.values)

    keys = ["param", "date", "time", "step", "level", "gridType", "type"]
    for k in keys:
        ds2.metadata(k) == ds.metadata(k)

    r = ds2.sel(level=850)
    assert len(r) == 1


@pytest.mark.parametrize("fl_type", ["file"])
def test_grib_serialise_file_fieldlist_concat(fl_type):
    ds00, _ = load_grib_data("test.grib", fl_type)
    ds01, _ = load_grib_data("test6.grib", fl_type)
    ds = ds00 + ds01
    assert len(ds) == 8

    pickled_f = pickle.dumps(ds)
    ds2 = pickle.loads(pickled_f)

    assert len(ds2) == 8
    assert np.allclose(ds[:2].values, ds2[:2].values)
    assert np.allclose(ds[2:].values, ds2[2:].values)

    keys = ["param", "date", "time", "step", "level", "gridType", "type"]
    for k in keys:
        ds2.metadata(k) == ds.metadata(k)


def test_grib_serialise_stream_1():
    with open(earthkit_examples_file("test.grib"), "rb") as f:
        ds = from_source("stream", f)
        with pytest.raises(NotImplementedError):
            pickle.dumps(ds)


def test_grib_serialise_stream_2():
    with open(earthkit_examples_file("test.grib"), "rb") as f:
        ds = from_source("stream", f, read_all=True)
        pickled_f = pickle.dumps(ds)

    ds2 = pickle.loads(pickled_f)

    assert len(ds2) == 2
    assert ds.metadata("shortName") == ["2t", "msl"]


def test_grib_serialise_file_parts():
    parts = (240, 150)

    ds = from_source("file", earthkit_examples_file("test6.grib"), parts=parts)
    assert len(ds) == 1

    pickled_f = pickle.dumps(ds)
    ds2 = pickle.loads(pickled_f)

    assert len(ds2) == 1
    assert ds2[0].metadata(["param", "level"]) == ["u", 1000]


@pytest.mark.parametrize("fl_type", FL_FILE)
@pytest.mark.parametrize("representation", ["file", "memory"])
@pytest.mark.parametrize("policy", ["path", "memory"])
def test_grib_serialise_policy(fl_type, representation, policy):
    ds, _ = load_grib_data("test.grib", fl_type)

    with config.temporary({"grib-file-serialisation-policy": policy}):
        ds2 = _pickle(ds, representation)

        assert len(ds2) == len(ds)
        assert ds2.values.shape == ds.values.shape
        if policy == "path":
            assert ds2.path == ds.path
        else:
            assert ds2.path != ds.path



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import pytest


@pytest.fixture
def hive_fs_1(fs):
    pattern = "/my_root/{shortName}_{date:date(%Y-%m-%d)}+{step}.grib"
    values = {
        "shortName": ["t", "z", "r"],
        "date": [datetime.datetime(2020, 5, 11), datetime.datetime(2020, 5, 12)],
        "step": [0, 6, 12],
    }
    sample = "/my_root/r_2020-05-11+6.grib"
    num = 18

    return build_hive_fs(pattern, values, sample, num, fs)


@pytest.fixture
def hive_fs_2(fs):
    pattern = "/my_root/{shortName}/{date:date(%Y-%m-%d)}:{step}.grib"
    values = {
        "shortName": ["t", "z", "r"],
        "date": [datetime.datetime(2020, 5, 11), datetime.datetime(2020, 5, 12)],
        "step": [0, 6, 12],
    }
    sample = "/my_root/r/2020-05-11:6.grib"
    num = 18

    return build_hive_fs(pattern, values, sample, num, fs)


@pytest.fixture
def hive_fs_3(fs):
    pattern = "/my_root/{step}__/{shortName}/my+{date:date(%Y-%m-%d)}.grib"
    values = {
        "shortName": ["t", "z", "r"],
        "date": [datetime.datetime(2020, 5, 11), datetime.datetime(2020, 5, 12)],
        "step": [0, 6, 12],
    }
    sample = "/my_root/6__/r/my+2020-05-11.grib"
    num = 18

    return build_hive_fs(pattern, values, sample, num, fs)


@pytest.fixture
def hive_fs_4(fs):
    pattern = "/my_root/{step}__/{shortName}/{shortName}_{date:date(%Y-%m-%d)}.grib"
    values = {
        "shortName": ["t", "z", "r"],
        "date": [datetime.datetime(2020, 5, 11), datetime.datetime(2020, 5, 12)],
        "step": [0, 6, 12],
    }
    sample = "/my_root/6__/r/r_2020-05-11.grib"
    num = 18

    return build_hive_fs(pattern, values, sample, num, fs)


@pytest.fixture
def hive_fs_5(fs):
    pattern = "/my_root/{step}/{shortName}/{date:date(%Y-%m-%d)}.grib"
    values = {
        "shortName": "t",
        "date": [datetime.datetime(2020, 5, 11), datetime.datetime(2020, 5, 12)],
        "step": [0, 6, 12],
    }
    sample = "/my_root/0/t/2020-05-11.grib"
    num = 6

    return build_hive_fs(pattern, values, sample, num, fs)


def build_hive_fs(pattern, values, sample, file_num, fs):
    from earthkit.data.utils.patterns import Pattern

    p = Pattern(pattern)
    files = p.substitute(values)
    assert len(files) == file_num
    assert sample in files

    for f in files:
        fs.create_file(f)

    return pattern, files, values



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest
from earthkit.utils.testing import check_array_type

from earthkit.data.testing import check_array

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from grib_fixtures import FL_FILE  # noqa: E402
from grib_fixtures import FL_NUMPY  # noqa: E402
from grib_fixtures import FL_TYPES  # noqa: E402
from grib_fixtures import load_grib_data  # noqa: E402


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_values_1(fl_type):
    f, array_backend = load_grib_data("test_single.grib", fl_type, folder="data")
    eps = 1e-5

    # whole file
    v = f.values
    check_array_type(v, array_backend, dtype="float64")
    assert v.shape == (1, 84)
    v = v[0].flatten()
    check_array(
        v,
        (84,),
        first=260.43560791015625,
        last=227.18560791015625,
        meanv=274.36566743396577,
        eps=eps,
        array_backend=array_backend,
    )

    # field
    v1 = f[0].values

    check_array_type(v1, array_backend)
    assert v1.shape == (84,)
    assert array_backend.allclose(v, v1, eps)


@pytest.mark.parametrize("fl_type", FL_FILE)
def test_grib_values_18(fl_type):
    f, array_backend = load_grib_data("tuv_pl.grib", fl_type)
    eps = 1e-5

    # whole file
    v = f.values
    check_array_type(v, array_backend, dtype="float64")
    assert v.shape == (18, 84)
    vf = v[0].flatten()
    check_array(
        vf,
        (84,),
        first=272.5642,
        last=240.56417846679688,
        meanv=279.70703560965404,
        eps=eps,
    )

    vf = v[15].flatten()
    check_array(
        vf,
        (84,),
        first=226.6531524658203,
        last=206.6531524658203,
        meanv=227.84362865629652,
        eps=eps,
    )


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_to_numpy_1(fl_type):
    f, _ = load_grib_data("test_single.grib", fl_type, folder="data")

    eps = 1e-5
    v = f.to_numpy()
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    v = v[0].flatten()
    check_array(
        v,
        (84,),
        first=260.43560791015625,
        last=227.18560791015625,
        meanv=274.36566743396577,
        eps=eps,
    )


@pytest.mark.parametrize("fl_type", FL_TYPES)
@pytest.mark.parametrize(
    "first,options, expected_shape",
    [
        (False, {}, (1, 7, 12)),
        (False, {"flatten": True}, (1, 84)),
        (False, {"flatten": False}, (1, 7, 12)),
        (True, {}, (7, 12)),
        (True, {"flatten": True}, (84,)),
        (True, {"flatten": False}, (7, 12)),
    ],
)
def test_grib_to_numpy_1_shape(fl_type, first, options, expected_shape):
    f, _ = load_grib_data("test_single.grib", fl_type, folder="data")

    v_ref = f[0].to_numpy().flatten()
    eps = 1e-5

    data = f[0] if first else f
    v1 = data.to_numpy(**options)
    assert isinstance(v1, np.ndarray)
    assert v1.dtype == np.float64
    assert v1.shape == expected_shape
    v1 = v1.flatten()
    assert np.allclose(v_ref, v1, eps)


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_to_numpy_18(fl_type):
    f, _ = load_grib_data("tuv_pl.grib", fl_type)

    eps = 1e-5

    # whole file
    v = f.to_numpy(flatten=True)
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (18, 84)
    vf0 = v[0].flatten()
    check_array(
        vf0,
        (84,),
        first=272.5642,
        last=240.56417846679688,
        meanv=279.70703560965404,
        eps=eps,
    )

    vf15 = v[15].flatten()
    check_array(
        vf15,
        (84,),
        first=226.6531524658203,
        last=206.6531524658203,
        meanv=227.84362865629652,
        eps=eps,
    )


@pytest.mark.parametrize("fl_type", FL_TYPES)
@pytest.mark.parametrize(
    "options, expected_shape",
    [
        (
            {},
            (
                18,
                7,
                12,
            ),
        ),
        (
            {"flatten": True},
            (
                18,
                84,
            ),
        ),
        ({"flatten": False}, (18, 7, 12)),
    ],
)
def test_grib_to_numpy_18_shape(fl_type, options, expected_shape):
    f, _ = load_grib_data("tuv_pl.grib", fl_type)

    eps = 1e-5

    # whole file
    v = f.to_numpy()
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (18, 7, 12)
    vf0 = f[0].to_numpy().flatten()
    assert vf0.shape == (84,)
    vf15 = f[15].to_numpy().flatten()
    assert vf15.shape == (84,)

    v1 = f.to_numpy(**options)
    assert isinstance(v1, np.ndarray)
    assert v1.dtype == np.float64
    assert v1.shape == expected_shape
    vr = v1[0].flatten()
    assert np.allclose(vf0, vr, eps)
    vr = v1[15].flatten()
    assert np.allclose(vf15, vr, eps)


@pytest.mark.parametrize("fl_type", FL_NUMPY)
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_grib_to_numpy_1_dtype(fl_type, dtype):
    f, _ = load_grib_data("test_single.grib", fl_type, folder="data")

    v = f[0].to_numpy(dtype=dtype)
    assert v.dtype == dtype

    v = f.to_numpy(dtype=dtype)
    assert v.dtype == dtype


@pytest.mark.parametrize("fl_type", FL_NUMPY)
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_grib_to_numpy_18_dtype(fl_type, dtype):
    f, _ = load_grib_data("tuv_pl.grib", fl_type)

    v = f[0].to_numpy(dtype=dtype)
    assert v.dtype == dtype

    v = f.to_numpy(dtype=dtype)
    assert v.dtype == dtype


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_to_numpy_1_index(fl_type):
    ds, _ = load_grib_data("test_single.grib", fl_type, folder="data")

    eps = 1e-5

    v = ds[0].to_numpy(flatten=True, index=[0, -1])
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (2,)
    assert np.allclose(v, [260.43560791015625, 227.18560791015625])

    v = ds[0].to_numpy(flatten=True, index=slice(None, None))
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    check_array(
        v,
        (84,),
        first=260.43560791015625,
        last=227.18560791015625,
        meanv=274.36566743396577,
        eps=eps,
    )

    v = ds[0].to_numpy(index=(slice(None, 2), slice(None, 3)))
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (2, 3)
    assert np.allclose(
        v,
        [
            [260.43560791, 260.43560791, 260.43560791],
            [280.81060791, 277.06060791, 284.43560791],
        ],
    )


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_to_numpy_18_index(fl_type):
    ds, _ = load_grib_data("tuv_pl.grib", fl_type)

    eps = 1e-5

    v = ds.to_numpy(flatten=True, index=[0, -1])
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (18, 2)
    vf0 = v[0].flatten()
    assert np.allclose(vf0, [272.5642, 240.56417846679688])
    vf15 = v[15].flatten()
    assert np.allclose(vf15, [226.6531524658203, 206.6531524658203])

    v = ds.to_numpy(flatten=True, index=slice(None, 2))
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (18, 2)
    vf0 = v[0].flatten()
    assert np.allclose(vf0, [272.56417847, 272.56417847])
    vf15 = v[15].flatten()
    assert np.allclose(vf15, [226.65315247, 226.65315247])

    v = ds.to_numpy(flatten=True, index=slice(None, None))
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (18, 84)
    vf0 = v[0].flatten()
    check_array(
        vf0,
        (84,),
        first=272.5642,
        last=240.56417846679688,
        meanv=279.70703560965404,
        eps=eps,
    )

    vf15 = v[15].flatten()
    check_array(
        vf15,
        (84,),
        first=226.6531524658203,
        last=206.6531524658203,
        meanv=227.84362865629652,
        eps=eps,
    )

    v = ds.to_numpy(index=(slice(None, 2), slice(None, 3)))
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (18, 2, 3)
    vf0 = v[0].flatten()
    assert np.allclose(
        vf0,
        [
            272.56417847,
            272.56417847,
            272.56417847,
            288.56417847,
            296.56417847,
            288.56417847,
        ],
    )
    vf15 = v[15].flatten()
    assert np.allclose(
        vf15,
        [
            226.65315247,
            226.65315247,
            226.65315247,
            230.65315247,
            230.65315247,
            230.65315247,
        ],
    )


@pytest.mark.parametrize("fl_type", FL_FILE)
# @pytest.mark.parametrize("fl_type", FL_NUMPY)
@pytest.mark.parametrize(
    "kwarg,expected_shape,expected_dtype",
    [
        ({}, (11, 19), np.float64),
        # ({"flatten": True}, (209,), np.float64),
        # ({"flatten": True, "dtype": np.float32}, (209,), np.float32),
        # ({"flatten": True, "dtype": np.float64}, (209,), np.float64),
        # ({"flatten": False}, (11, 19), np.float64),
        # ({"flatten": False, "dtype": np.float32}, (11, 19), np.float32),
        # ({"flatten": False, "dtype": np.float64}, (11, 19), np.float64),
    ],
)
def test_grib_field_data(fl_type, kwarg, expected_shape, expected_dtype):
    ds, _ = load_grib_data("test.grib", fl_type)

    latlon = ds[0].to_latlon(**kwarg)
    v = ds[0].to_numpy(**kwarg)

    d = ds[0].data(**kwarg)
    assert isinstance(d, np.ndarray)
    assert d.dtype == expected_dtype
    assert len(d) == 3
    assert d[0].shape == expected_shape
    assert np.allclose(d[0], latlon["lat"])
    assert np.allclose(d[1], latlon["lon"])
    assert np.allclose(d[2], v)

    d = ds[0].data(keys="lat", **kwarg)
    assert d.shape == expected_shape
    assert d.dtype == expected_dtype
    assert np.allclose(d, latlon["lat"])

    d = ds[0].data(keys="lon", **kwarg)
    assert d.shape == expected_shape
    assert d.dtype == expected_dtype
    assert np.allclose(d, latlon["lon"])

    d = ds[0].data(keys="value", **kwarg)
    assert d.shape == expected_shape
    assert d.dtype == expected_dtype
    assert np.allclose(d, v)

    d = ds[0].data(keys=("value", "lon"), **kwarg)
    assert isinstance(d, np.ndarray)
    assert d.dtype == expected_dtype
    assert len(d) == 2
    assert np.allclose(d[0], v)
    assert np.allclose(d[1], latlon["lon"])


@pytest.mark.parametrize("fl_type", FL_NUMPY)
@pytest.mark.parametrize(
    "kwarg,expected_shape,expected_dtype",
    [
        ({}, (11, 19), np.float64),
        ({"flatten": True}, (209,), np.float64),
        ({"flatten": True, "dtype": np.float32}, (209,), np.float32),
        ({"flatten": True, "dtype": np.float64}, (209,), np.float64),
        ({"flatten": False}, (11, 19), np.float64),
        ({"flatten": False, "dtype": np.float32}, (11, 19), np.float32),
        ({"flatten": False, "dtype": np.float64}, (11, 19), np.float64),
    ],
)
def test_grib_fieldlist_data(fl_type, kwarg, expected_shape, expected_dtype):
    ds, _ = load_grib_data("test.grib", fl_type)

    latlon = ds.to_latlon(**kwarg)
    v = ds.to_numpy(**kwarg)

    d = ds.data(**kwarg)
    assert isinstance(d, np.ndarray)
    assert d.shape == tuple([4, *expected_shape])
    assert d.dtype == expected_dtype
    assert np.allclose(d[0], latlon["lat"])
    assert np.allclose(d[1], latlon["lon"])
    assert np.allclose(d[2], v[0])
    assert np.allclose(d[3], v[1])

    d = ds.data(keys="lat", **kwarg)
    assert d.shape == tuple([1, *expected_shape])
    assert d.dtype == expected_dtype
    assert np.allclose(d[0], latlon["lat"])

    d = ds.data(keys="lon", **kwarg)
    assert d.shape == tuple([1, *expected_shape])
    assert d.dtype == expected_dtype
    assert np.allclose(d[0], latlon["lon"])

    d = ds.data(keys="value", **kwarg)
    assert d.shape == tuple([2, *expected_shape])
    assert d.dtype == expected_dtype
    assert np.allclose(d, v)

    d = ds.data(keys=("value", "lon"), **kwarg)
    assert isinstance(d, np.ndarray)
    assert d.shape == tuple([3, *expected_shape])
    assert d.dtype == expected_dtype
    assert np.allclose(d[0], v[0])
    assert np.allclose(d[1], v[1])
    assert np.allclose(d[2], latlon["lon"])


@pytest.mark.parametrize("fl_type", FL_NUMPY)
def test_grib_fieldlist_data_index(fl_type):
    ds, _ = load_grib_data("tuv_pl.grib", fl_type)

    eps = 1e-5

    latlon = ds.to_latlon(flatten=True)
    lat = latlon["lat"]
    lon = latlon["lon"]

    index = [0, -1]
    v = ds.data(flatten=True, index=index)
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (18 + 2, 2)
    assert np.allclose(v[0].flatten(), lat[index])
    assert np.allclose(v[1].flatten(), lon[index])
    vf0 = v[2 + 0].flatten()
    assert np.allclose(vf0, [272.5642, 240.56417846679688])
    vf15 = v[2 + 15].flatten()
    assert np.allclose(vf15, [226.6531524658203, 206.6531524658203])

    index = slice(None, 2)
    v = ds.data(flatten=True, index=index)
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (18 + 2, 2)
    assert np.allclose(v[0].flatten(), lat[index])
    assert np.allclose(v[1].flatten(), lon[index])
    vf0 = v[2 + 0].flatten()
    assert np.allclose(vf0, [272.56417847, 272.56417847])
    vf15 = v[2 + 15].flatten()
    assert np.allclose(vf15, [226.65315247, 226.65315247])

    index = slice(None, None)
    v = ds.data(flatten=True, index=index)
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (18 + 2, 84)
    assert np.allclose(v[0].flatten(), lat)
    assert np.allclose(v[1].flatten(), lon)
    vf0 = v[2 + 0].flatten()
    check_array(
        vf0,
        (84,),
        first=272.5642,
        last=240.56417846679688,
        meanv=279.70703560965404,
        eps=eps,
    )

    vf15 = v[2 + 15].flatten()
    check_array(
        vf15,
        (84,),
        first=226.6531524658203,
        last=206.6531524658203,
        meanv=227.84362865629652,
        eps=eps,
    )

    index = (slice(None, 2), slice(None, 3))
    v = ds.data(index=index)
    assert isinstance(v, np.ndarray)
    assert v.dtype == np.float64
    assert v.shape == (2 + 18, 2, 3)
    latlon = ds.to_latlon()
    lat = latlon["lat"]
    lon = latlon["lon"]
    assert np.allclose(v[0], lat[index])
    assert np.allclose(v[1], lon[index])

    vf0 = v[2 + 0].flatten()
    assert np.allclose(
        vf0,
        [
            272.56417847,
            272.56417847,
            272.56417847,
            288.56417847,
            296.56417847,
            288.56417847,
        ],
    )
    vf15 = v[2 + 15].flatten()
    assert np.allclose(
        vf15,
        [
            226.65315247,
            226.65315247,
            226.65315247,
            230.65315247,
            230.65315247,
            230.65315247,
        ],
    )


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_values_with_missing(fl_type):
    f, array_backend = load_grib_data("test_single_with_missing.grib", fl_type, folder="data")

    v = f[0].values
    check_array_type(v, array_backend)
    assert v.shape == (84,)
    eps = 0.001

    ns = array_backend.namespace

    assert ns.count_nonzero(ns.isnan(v)) == 38
    mask = array_backend.from_other([12, 14, 15, 24, 25, 26] + list(range(28, 60)))
    v1 = array_backend.to_numpy(v)
    assert np.isclose(v1[0], 260.4356, eps)
    assert np.isclose(v1[11], 260.4356, eps)
    assert np.isclose(v1[-1], 227.1856, eps)
    m = v[mask]
    assert len(m) == 38
    assert ns.count_nonzero(ns.isnan(m)) == 38


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest

here = os.path.dirname(__file__)
sys.path.insert(0, here)

from lod_fixtures import build_lod_fieldlist  # noqa: E402


@pytest.mark.parametrize("mode", ["list-of-dicts", "loop"])
def test_lod_geo_distinct_ll(lod_distinct_ll, mode):
    ds = build_lod_fieldlist(lod_distinct_ll, mode)

    assert len(ds) == 6
    assert ds[0].shape == (3, 2)

    lat_ref = np.array([[-10.0, -10.0], [0.0, 0.0], [10.0, 10.0]])
    lon_ref = np.array([[20.0, 40.0], [20.0, 40.0], [20.0, 40.0]])

    ll = ds[0].to_latlon()
    lat = ll["lat"]
    lon = ll["lon"]
    assert np.allclose(lat, lat_ref)
    assert np.allclose(lon, lon_ref)

    gr = ds[0].mars_area
    assert isinstance(gr, list)
    assert np.allclose(np.array(gr), np.array([10, 20, -10, 40]))

    gr = ds[0].mars_grid
    assert isinstance(gr, list)
    assert np.allclose(np.array(gr), np.array([20.0, 10.0]))

    gr = ds[0].grid_points()
    assert len(gr) == 2
    assert np.allclose(gr[0], lat_ref.flatten())
    assert np.allclose(gr[1], lon_ref.flatten())

    assert ds[0].resolution is None

    ll = ds.to_latlon()
    lat = ll["lat"]
    lon = ll["lon"]
    assert np.allclose(lat, lat_ref)
    assert np.allclose(lon, lon_ref)


@pytest.mark.parametrize("mode", ["list-of-dicts", "loop"])
def test_lod_geo_ll_flat(lod_ll_flat, mode):
    ds = build_lod_fieldlist(lod_ll_flat, mode)

    assert len(ds) == 6
    assert ds[0].shape == (6,)

    lat_ref = np.array([-10.0, -10.0, 0.0, 0.0, 10.0, 10.0])
    lon_ref = np.array([20.0, 40.0, 20.0, 40.0, 20.0, 40.0])

    ll = ds[0].to_latlon()
    lat = ll["lat"]
    lon = ll["lon"]
    assert np.allclose(lat, lat_ref)
    assert np.allclose(lon, lon_ref)

    gr = ds[0].mars_area
    assert isinstance(gr, list)
    assert np.allclose(np.array(gr), np.array([10, 20, -10, 40]))

    with pytest.raises(NotImplementedError):
        gr = ds[0].mars_grid

    gr = ds[0].grid_points()
    assert len(gr) == 2
    assert np.allclose(gr[0], lat_ref.flatten())
    assert np.allclose(gr[1], lon_ref.flatten())

    assert ds[0].resolution is None

    ll = ds.to_latlon()
    lat = ll["lat"]
    lon = ll["lon"]
    assert np.allclose(lat, lat_ref)
    assert np.allclose(lon, lon_ref)


@pytest.mark.parametrize("data", ["lod_ll_2D_all", "lod_ll_2D_ll", "lod_ll_2D_values"])
@pytest.mark.parametrize("mode", ["list-of-dicts", "loop"])
def test_lod_geo_ll_2D(request, data, mode):
    ds = build_lod_fieldlist(request.getfixturevalue(data), mode)

    assert len(ds) == 6
    assert ds[0].shape == (3, 2)

    lat_ref = np.array([[-10.0, -10.0], [0.0, 0.0], [10.0, 10.0]])
    lon_ref = np.array([[20.0, 40.0], [20.0, 40.0], [20.0, 40.0]])

    ll = ds[0].to_latlon()
    lat = ll["lat"]
    lon = ll["lon"]
    assert np.allclose(lat, lat_ref)
    assert np.allclose(lon, lon_ref)

    gr = ds[0].mars_area
    assert isinstance(gr, list)
    assert np.allclose(np.array(gr), np.array([10, 20, -10, 40]))

    # gr = ds[0].mars_grid
    # assert isinstance(gr, list)
    # assert np.allclose(np.array(gr), np.array([20.0, 10.0]))

    gr = ds[0].grid_points()
    assert len(gr) == 2
    assert np.allclose(gr[0], lat_ref.flatten())
    assert np.allclose(gr[1], lon_ref.flatten())

    assert ds[0].resolution is None

    ll = ds.to_latlon()
    lat = ll["lat"]
    lon = ll["lon"]
    assert np.allclose(lat, lat_ref)
    assert np.allclose(lon, lon_ref)


@pytest.mark.parametrize("data", ["lod_ll_flat_invalid"])
@pytest.mark.parametrize("mode", ["list-of-dicts", "loop"])
def test_lod_geo_invalid(
    request,
    data,
    mode,
):
    ds = build_lod_fieldlist(request.getfixturevalue(data), mode)

    assert len(ds) == 6
    with pytest.raises(ValueError):
        ds[0].shape


@pytest.mark.parametrize("data", ["lod_no_latlon"])
@pytest.mark.parametrize("mode", ["list-of-dicts", "loop"])
def test_lod_no_latlon(
    request,
    data,
    mode,
):
    ds = build_lod_fieldlist(request.getfixturevalue(data), mode)

    assert len(ds) == 6
    assert ds[0].shape == (6,)
    assert ds[0].values.shape == (6,)
    assert ds.values.shape == (6, 6)

    with pytest.raises(ValueError):
        ds[0].to_latlon()


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



import earthkit.data as ekd

ds = ekd.from_source("sample", "test.grib")

# case 1
o = ekd.new_grib_output("res_{shortName}.grib", split_output=True)
for field in ds:
    o.write(None, template=field)
o.close()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import re

from earthkit.data.core.config import CONFIG_AND_HELP
from earthkit.data.core.config import env_var_name


def execute(*args):
    print()
    print(".. list-table::")
    print("   :header-rows: 1")
    print("   :widths: 40 60")
    print()
    print("   * - Config option name")
    print("     - Environment variable")
    print()
    for k, v in sorted(CONFIG_AND_HELP.items()):
        if len(args) and not any(re.match(arg, k) for arg in args):
            continue

        print("   * -  ", k.replace("-", "\u2011"))  # Non-breaking hyphen
        print("     -  ", f"{env_var_name(k)}")
    print()


if __name__ == "__main__":
    execute()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from abc import ABCMeta
from abc import abstractmethod


class Geography(metaclass=ABCMeta):
    """Geographical information about a field or data unit"""

    @abstractmethod
    def latitudes(self, dtype=None):
        r"""Return the latitudes.

        Parameters
        ----------
        dtype: str, numpy.dtype or None
            Typecode or data-type of the array. When it is :obj:`None` the default
            type used by the underlying data accessor is used. For GRIB it is
            ``np.float64``.

        Returns
        -------
        ndarray
        """
        pass

    @abstractmethod
    def longitudes(self, dtype=None):
        r"""Return the longitudes.

        Parameters
        ----------
        dtype: str, numpy.dtype or None
            Typecode or data-type of the array. When it is :obj:`None` the default
            type used by the underlying data accessor is used. For GRIB it is
            ``np.float64``.

        Returns
        -------
        ndarray
        """
        pass

    @abstractmethod
    def x(self, dtype=None):
        r"""Return the x coordinates in the original CRS.

        Parameters
        ----------
        dtype: str, numpy.dtype or None
            Typecode or data-type of the array. When it is :obj:`None` the default
            type used by the underlying data accessor is used. For GRIB it is
            ``np.float64``.

        Returns
        -------
        ndarray
        """
        pass

    @abstractmethod
    def y(self, dtype=None):
        r"""Return the y coordinates in the original CRS.

        Parameters
        ----------
        dtype: str, numpy.dtype or None
            Typecode or data-type of the array. When it is :obj:`None` the default
            type used by the underlying data accessor is used. For GRIB it is
            ``np.float64``.

        Returns
        -------
        ndarray
        """
        pass

    @abstractmethod
    def shape(self):
        r"""Return the shape of the grid or data values.

        Returns
        -------
        tuple
        """
        pass

    @abstractmethod
    def _unique_grid_id(self):
        r"""Return a unique id of the grid of a field."""
        pass

    @abstractmethod
    def projection(self):
        r"""Return information about projection.

        Returns
        -------
        :obj:`Projection`
        """
        pass

    @abstractmethod
    def bounding_box(self):
        r"""Return the bounding box.

        Returns
        -------
        :obj:`BoundingBox <data.utils.bbox.BoundingBox>`
        """
        pass

    @abstractmethod
    def gridspec(self):
        r"""Return the grid specification.

        Returns
        -------
        :class:`~data.core.gridspec.GridSpec>`
        """
        pass

    @abstractmethod
    def resolution(self):
        pass

    @abstractmethod
    def mars_grid(self):
        pass

    @abstractmethod
    def mars_area(self):
        pass



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging
from itertools import product

from earthkit.data.utils.dates import date_to_grib
from earthkit.data.utils.dates import datetime_to_grib
from earthkit.data.utils.dates import step_to_grib
from earthkit.data.utils.dates import time_to_grib
from earthkit.data.utils.dates import to_datetime

LOG = logging.getLogger(__name__)


def update_metadata(metadata, compulsory):
    if "valid_time" in metadata:
        dt = to_datetime(metadata.pop("valid_time"))
        metadata["date"] = dt.date()
        metadata["time"] = dt.time()

    if "forecast_reference_time" in metadata:
        date, time = datetime_to_grib(to_datetime(metadata["forecast_reference_time"]))
        metadata.pop("forecast_reference_time")
        metadata["date"] = date
        metadata["time"] = time

    if "time" in metadata:
        metadata["time"] = time_to_grib(metadata["time"])

    if "date" in metadata:
        date = metadata["date"]
        metadata["date"] = date_to_grib(date)
        if "time" not in metadata:
            if isinstance(date, datetime.datetime):
                metadata["time"] = date.hour * 100 + date.minute

    if "step" in metadata:
        metadata["step"] = step_to_grib(metadata["step"])

    if "stream" not in metadata:
        if "number" in metadata:
            metadata["stream"] = "enfo"
            metadata.setdefault("type", "pf")

    if "levelist" in metadata:
        metadata.setdefault("levtype", "pl")

    if "param" in metadata:
        param = metadata.pop("param")
        try:
            metadata["paramId"] = int(param)
        except ValueError:
            metadata["shortName"] = param

    # levtype is a readOnly key in ecCodes >= 2.33.0
    levtype_remap = {
        "pl": "isobaricInhPa",
        "ml": "hybrid",
        "pt": "theta",
        "pv": "potentialVorticity",
        "sfc": "surface",
    }
    if "levtype" in metadata:
        v = metadata.pop("levtype")
        metadata["typeOfLevel"] = levtype_remap[v]


def coord_to_component(coord):
    if "_earthkit" in coord.attrs:
        keys = coord.attrs["_earthkit"].get("keys", [])
        values = coord.attrs["_earthkit"].get("values", [])

        # for datetime composite coords only the keys are added as an attribute
        if not values and len(coord) == 2:
            values = [datetime_to_grib(to_datetime(x)) for x in coord.values]

        if len(coord) == len(values):
            r = [[a, *b] for a, b in zip(coord.values, values)]
            return r, keys
        else:
            raise ValueError(f"Cannot extract components for coordinate {coord.name}")
    return None


def data_array_to_fields(da, metadata=None):
    from earthkit.data.sources.array_list import ArrayField

    dims = [dim for dim in da.dims if dim not in ["values", "X", "Y", "lat", "lon", "latitude", "longitude"]]
    coords = {k: v for k, v in da.coords.items() if k in dims}
    components = {}
    for k in coords:
        c = coord_to_component(coords[k])
        if c:
            coords[k] = c[0]
            components[k] = c[1]
        else:
            coords[k] = coords[k].values

    # print(f"{coords=}")

    for values in product(*[coords[dim] for dim in dims]):

        # field
        local_coords = dict(zip(dims, values))
        for k in components:
            local_coords[k] = local_coords[k][0]

        # print("local_coords", local_coords)
        xa_field = da.sel(**local_coords)

        # metadata
        grib_metadata = dict(zip(dims, values))
        for k in components:
            # print(f"{k=}")
            # print(grib_metadata[k])
            grib_metadata.update(dict(zip(components[k], grib_metadata[k][1:])))
            # print(f"-> {grib_metadata=}")
            del grib_metadata[k]
        update_metadata(grib_metadata, [])

        # extract metadata from object
        if metadata is None:
            if hasattr(da, "earthkit"):
                metadata = da.earthkit.metadata
            else:
                raise ValueError(
                    "Earthkit attribute not found in DataArray. Required for conversion to FieldList!"
                )

        metadata = metadata.override(grib_metadata)
        yield ArrayField(xa_field.values, metadata)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_object
from earthkit.data import wrappers
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.wrappers import xarray as xr_wrapper


@pytest.mark.no_eccodes
def test_dataset_wrapper():
    import xarray as xr

    _wrapper = xr_wrapper.wrapper(xr.Dataset())
    assert isinstance(_wrapper, xr_wrapper.XArrayDatasetWrapper)
    _wrapper = wrappers.get_wrapper(xr.Dataset())
    assert isinstance(_wrapper, xr_wrapper.XArrayDatasetWrapper)
    _wrapper = from_object(xr.Dataset())
    assert isinstance(_wrapper, xr_wrapper.XArrayDatasetWrapper)


@pytest.mark.no_eccodes
def test_dataarray_wrapper():
    import xarray as xr

    _wrapper = xr_wrapper.wrapper(xr.DataArray())
    assert isinstance(_wrapper, xr_wrapper.XArrayDataArrayWrapper)
    _wrapper = wrappers.get_wrapper(xr.DataArray())
    assert isinstance(_wrapper, xr_wrapper.XArrayDataArrayWrapper)
    _wrapper = from_object(xr.DataArray())
    assert isinstance(_wrapper, xr_wrapper.XArrayDataArrayWrapper)


@pytest.mark.no_eccodes
def test_xarray_lazy_fieldlist_scan():
    import xarray as xr

    ds = from_object(xr.open_dataset(earthkit_examples_file("test.nc")))
    assert ds._fields is None
    assert len(ds) == 2
    assert len(ds._fields) == 2


@pytest.mark.no_eccodes
def test_xarray_to_numpy():
    import xarray as xr

    data_array = xr.DataArray(
        [[1, 2, 3], [4, 5, 6]],
        dims=["x", "y"],
        coords={"x": [1, 2], "y": [3, 4, 5]},
    )
    ds = from_object(data_array)

    arr_2d = ds.to_numpy()
    assert arr_2d.shape == (2, 3)

    arr_1d = ds.to_numpy(flatten=True)
    assert arr_1d.shape == (6,)



import earthkit.data as ekd

ds = ekd.from_source("sample", "test.grib")

# case 1
ds.to_target("file", "res.grib")

# case 2
ds.to_target("file", "res.grib", append=True)

# case 3
ds.to_target("file", "res.grib", metadata={"bitsPerValue": 12})

# case 4 - equivalent solutions
with ekd.create_target("file", "res1.grib") as t:
    for field in ds:
        t.write(field, metadata={"bitsPerValue": 12})

with ekd.create_target("file", "res1.grib", metadata={"bitsPerValue": 12}) as t:
    for field in ds:
        t.write(field)

with ekd.create_target("file", "res1.grib") as t:
    for field in ds:
        t.write(field, bitsPerValue=12)



import earthkit.data as ekd

ds = ekd.from_source("sample", "test.grib")

# case 1
ds.save("res.grib")

# case 2
ds.save("res.grib", append=True)

# case 3
ds.save("res.grib", bits_per_value=12)

# case 4
for field in ds:
    field.save("res1.grib", append=True, bits_per_value=12)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from earthkit.data.decorators import normalize


def test_param_convention_mars_1():
    @normalize("parameter", "variable-list", convention="mars")
    def values_mars(parameter):
        return parameter

    assert values_mars(parameter="tp") == ["tp"]
    assert values_mars(parameter="2t") == ["2t"]
    assert values_mars(parameter="t2m") == ["2t"]
    assert values_mars(parameter=["t2m", "tp"]) == ["2t", "tp"]
    assert values_mars(parameter="whatever") == ["whatever"]


def test_param_convention_mars_2():
    @normalize("parameter", "variable-list(mars)")
    def values_mars(parameter):
        return parameter

    assert values_mars(parameter="tp") == ["tp"]
    assert values_mars(parameter="2t") == ["2t"]
    assert values_mars(parameter="t2m") == ["2t"]
    assert values_mars(parameter=["t2m", "tp"]) == ["2t", "tp"]
    assert values_mars(parameter="whatever") == ["whatever"]


def test_param_convention_cf_list():
    @normalize("parameter", "variable-list(cf)")
    def values_cf(parameter):
        return parameter

    assert values_cf(parameter="tp") == ["tp"]
    assert values_cf(parameter="2t") == ["t2m"]
    assert values_cf(parameter="t2m") == ["t2m"]


def test_param_convention_cf():
    @normalize("parameter", "variable(cf)")
    def values_cf(parameter):
        return parameter

    assert values_cf(parameter="tp") == "tp"
    assert values_cf(parameter="2t") == "t2m"
    assert values_cf(parameter="t2m") == "t2m"


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import functools
import itertools
import logging
from abc import ABCMeta
from abc import abstractmethod

import numpy as np

from earthkit.data.core.index import Selection
from earthkit.data.core.index import normalize_selection

LOG = logging.getLogger(__name__)


def coords_to_index(coords, shape) -> int:
    """
    Map user coords to field index"""
    index = 0
    n = 1
    for i in range(len(coords) - 1, -1, -1):
        index += coords[i] * n
        n *= shape[i]
    return index


def index_to_coords(index: int, shape):
    assert isinstance(index, int), (index, type(index))

    result = [None] * len(shape)
    i = len(shape) - 1

    while i >= 0:
        result[i] = index % shape[i]
        index = index // shape[i]
        i -= 1

    result = tuple(result)

    assert len(result) == len(shape)
    return result


class CubeSelection(Selection):
    def match_element(self, element):
        return all(v(element) for k, v in self.actions.items())


class CubeCoords(dict):
    def __repr__(self):
        t = "Coordinates:\n"
        max_len = max(len(k) for k in self.keys())
        for k, v in self.items():
            t += f"  {k:<{max_len+4}}{self._format_item(v)}\n"
        return t

    def _format_item(self, item, n=10):
        if len(item) == 0:
            return "??"

        t = f"[{type(item[0]).__name__}] "
        if len(item) == 1:
            t += str(item[0])
        elif len(item) < n:
            t += ", ".join([str(s) for s in item])
        else:
            t += ", ".join([str(s) for s in item[: n - 1]]) + " ,..., " + str(item[-1])

        return t


def flatten_arg(func):
    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        _kwargs = {**kwargs}
        _kwargs["flatten"] = len(self.field_shape) == 1
        return func(self, *args, **_kwargs)

    return wrapped


class TensorCore(metaclass=ABCMeta):
    _type = None
    _shape = None
    _user_coords = None
    _user_dims = None
    _user_shape = None
    _field_coords = None
    _field_dims = None
    _field_shape = None
    _array = None
    _data = None
    flatten_values = None

    @property
    def full_shape(self):
        return self._full_shape

    @property
    def user_shape(self):
        return self._user_shape

    @property
    def field_shape(self):
        return self._field_shape

    @property
    def full_dims(self):
        d = dict(self._user_dims)
        d.update(self._field_dims)
        return d

    @property
    def user_dims(self):
        return self._user_dims

    @property
    def field_dims(self):
        return self._field_dims

    @property
    def full_coords(self):
        d = dict(self._user_coords)
        d.update(self._field_coords)
        return d

    @property
    def user_coords(self):
        return self._user_coords

    @property
    def field_coords(self):
        return self._field_coords

    @abstractmethod
    def to_numpy(self, **kwargs):
        pass

    def __getitem__(self, indexes):
        # Make sure the requested indexes are a list of slices matching the shape

        if isinstance(indexes, int):
            indexes = (indexes,)  # make tuple

        if isinstance(indexes, list):
            indexes = tuple(indexes)

        assert isinstance(indexes, tuple), (type(indexes), indexes)

        indexes = list(indexes)

        if indexes[-1] is Ellipsis:
            indexes.pop()

        while len(indexes) <= len(self.user_shape):
            indexes.append(slice(None, None, None))

        while len(indexes) > len(self.user_shape):
            indexes.pop()

        # print(f"{indexes=} user_shape={self.user_shape=}")
        indexes = tuple(indexes)

        return self._subset(indexes)

    def sel(self, *args, remapping=None, **kwargs):
        kwargs = normalize_selection(*args, **kwargs)

        r = {}
        for k, v in kwargs.items():
            selection = CubeSelection(dict(k=v))
            r[k] = list(
                i for i, element in enumerate(self.user_coords[k]) if selection.match_element(element)
            )
            if len(r[k]) == 1:
                v = r[k][0]
                r[k] = slice(v, v + 1)
                # r[k] = r[k][0]

        indexes = []
        for k, v in self.user_coords.items():
            if k in r:
                indexes.append(r[k])
            else:
                indexes.append(slice(None, None, None))

        indexes = tuple(indexes)
        # print(f"{indexes=}")

        return self._subset(indexes)

    def isel(self, *args, remapping=None, **kwargs):
        # print("isel", args, kwargs)
        # print("isel", self.coords)
        kwargs = normalize_selection(*args, **kwargs)

        indexes = []
        for k, v in self.user_coords.items():
            if k in kwargs:
                indexes.append(kwargs[k])
            else:
                indexes.append(slice(None, None, None))

        indexes = tuple(indexes)
        # print(f"{indexes=}")

        return self._subset(indexes)

    @abstractmethod
    def _subset(self, indexes):
        pass

    def _subset_coords(self, indexes):
        import numpy as np

        # TODO: avoid copying values
        r = CubeCoords()
        for i, k in enumerate(self.user_coords.keys()):
            if i < len(indexes):
                idx = indexes[i]
                # print(f"{k=} {idx=} {self.coords[k]}")
                # print(self.coords[k][idx])
                if isinstance(idx, (int, slice)):
                    # print(f"{k=} {idx=} {self._user_coords[k]}")
                    v = self._user_coords[k][idx]
                    if not isinstance(v, (list, tuple, np.ndarray)):
                        v = (v,)
                    r[k] = v
                elif isinstance(idx, list):
                    r[k] = tuple([self._user_coords[k][i] for i in idx])
            else:
                r[k] = self._user_coords[k]
        return r

    @staticmethod
    def _index_to_coords_value(index, tensor):
        coord_idx = index_to_coords(index, tensor._user_shape)
        coords = []
        for k, v in enumerate(tensor._user_coords.values()):
            coords.append(v[coord_idx[k]])
        return coords

    def _check(self):
        if self._full_shape != self._user_shape + self._field_shape:
            raise ValueError(
                (
                    f"shape={self._full_shape} differs from expected shape="
                    f"{self._user_shape} + {self._field_shape}"
                )
            )

        shape = self._coords_shape(self._user_coords) + self._dims_shape(self._field_dims)
        if shape != self._full_shape:
            raise ValueError(f"shape={self._full_shape} does not match shape deduced from coords={shape}")

    # def copy(self, data=None):
    #     if data is None:
    #         data = self.to_numpy().copy()
    #     return ArrayTensor(data, self.user_coords, self.field_shape)

    @staticmethod
    def _coords_shape(coords):
        return tuple(len(v) for _, v in coords.items())

    @staticmethod
    def _dims_shape(dims):
        return tuple(v for _, v in dims.items())


class FieldListTensor(TensorCore):
    def __init__(
        self,
        source,
        user_coords,
        field_coords,
        field_dims,
        flatten_values,
    ):
        # print(f"FieldListTensor user_coords={user_coords}")
        # print(f"FieldListTensor field_coords={field_coords.keys()} {field_dims=}")

        self.source = source
        self._user_coords = user_coords
        self._user_shape = self._coords_shape(user_coords)
        self._user_dims = {k: len(v) for k, v in user_coords.items()}
        self._field_coords = field_coords
        self._field_shape = self._dims_shape(field_dims)
        self._field_dims = field_dims
        self._full_shape = self._user_shape + self._field_shape
        self.flatten_values = flatten_values

        # consistency check
        from earthkit.data.utils.xarray.check import CubeChecker

        checker = CubeChecker(self)
        checker.check(details=True)

    @classmethod
    def from_tensor(cls, owner, source, user_coords):
        return cls(
            source,
            user_coords,
            owner.field_coords,
            owner.field_dims,
            owner.flatten_values,
        )

    @classmethod
    def from_fieldlist(
        cls,
        ds,
        *args,
        remapping=None,
        flatten_values=False,
        sort=True,
        progress_bar=False,
        user_dims_and_coords=None,
        field_dims_and_coords=None,
    ):
        assert len(ds), f"No data in {ds}"

        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = args[0]

        names = []
        for a in args:
            if isinstance(a, str):
                names.append(a)
            elif isinstance(a, dict):
                names += list(a.keys())

        # Sort the source
        if names and sort:
            source = ds.order_by(*args, remapping=remapping)
        else:
            source = ds

        # Get a mapping of user names to unique values
        # With possible reduce dimensionality if the user uses 'level+param'
        if names:
            if user_dims_and_coords:
                user_coords = CubeCoords(user_dims_and_coords)
            else:
                user_coords = CubeCoords(
                    ds.unique_values(*names, remapping=remapping, progress_bar=progress_bar)
                )
                for k, v in user_coords.items():
                    user_coords[k] = tuple(sorted(v))
        else:
            user_coords = CubeCoords()

        # field properties
        if field_dims_and_coords is not None:
            field_dims, field_coords = field_dims_and_coords
        else:
            from earthkit.data.utils.xarray.grid import TensorGrid

            field_dims, field_coords, _ = TensorGrid.build(source[0], flatten_values)

        return cls(source, user_coords, field_coords, field_dims, flatten_values)

    def clear(self):
        self.source = None
        self._user_coords = None
        self._user_shape = None
        self._user_dims = None
        self._field_coords = None
        self._field_shape = None
        self._field_dims = None
        self._full_shape = None
        self.flatten_values = None

    @flatten_arg
    def to_numpy(self, index=None, **kwargs):
        if index is not None:
            if all(i == slice(None, None, None) for i in index):
                index = None

        if index is None:
            return self.source.to_numpy(**kwargs).reshape(*self.full_shape)
        else:
            n = self.source.to_numpy(index=index, **kwargs)
            shape = list(self._user_shape)
            shape += list(n.shape[1:])
            return n.reshape(*shape) if len(shape) > 0 else n

    @flatten_arg
    def to_array(self, index=None, **kwargs):
        if index is not None:
            if all(i == slice(None, None, None) for i in index):
                index = None

        if index is None:
            return self.source.to_array(**kwargs).reshape(*self.full_shape)
        else:
            n = self.source.to_array(index=index, **kwargs)
            shape = list(self._user_shape)
            shape += list(n.shape[1:])
            return n.reshape(*shape) if len(shape) > 0 else n

    @flatten_arg
    def latitudes(self, **kwargs):
        return self.source[0].data("lat", **kwargs)

    @flatten_arg
    def longitudes(self, **kwargs):
        return self.source[0].data("lon", **kwargs)

    def field_indexes(self, indexes):
        assert len(indexes) == len(self._full_shape)
        return indexes[len(self._user_shape) :]

    def is_full_field(self, indexes):
        assert len(indexes) == len(self._field_shape)
        for i, s in enumerate(indexes):
            if not (s is None or s == slice(None, None, None) or s == slice(0, self._field_shape[i], 1)):
                return False
        return True

    def _subset(self, indexes):
        """Only allow subsetting for the user coordinates.
        Indices for the field coordinates are ignored.
        """
        # Map the slices to a list of indexes per dimension
        assert len(indexes) >= len(self._user_shape)
        user_coords = []
        user_indexes = []

        for s, c in zip(indexes, self._user_shape):
            lst = np.array(list(range(c)))[s].tolist()
            if not isinstance(lst, list):
                lst = [lst]
            user_coords.append(lst)
            user_indexes.append(s)

        # print(f"{user_coords=} {user_indexes=}")

        assert len(user_coords) == len(self._user_coords)

        dataset_indexes = []
        user_shape = self._user_shape
        for x in itertools.product(*user_coords):
            i = coords_to_index(x, user_shape)
            assert isinstance(i, int), i
            dataset_indexes.append(i)

        coords = self._subset_coords(user_indexes)
        assert len(coords) == len(self._user_coords)
        ds = self.source[tuple(dataset_indexes)]
        return self.from_tensor(self, ds, coords)

    def make_valid_datetime(self, dims_map, dtype="datetime64[ns]"):
        # TODO: make it more general

        for k in ["valid_datetime", "valid_time"]:
            if k in self.user_coords:
                import datetime

                return (k,), [datetime.datetime.fromisoformat(x) for x in self.user_coords[k]]

        # in the tensor the dims.coords are GRIB keys
        # dims_map is a mapping from dim names to GRIB keys
        DIM_ROLES = {
            "forecast_reference_time": ("forecast_reference_time", "base_datetime"),
            "step": ("step_timedelta", "step", "ensStep", "stepRange"),
            "date": ("date", "dataDate"),
            "time": ("time", "dataTime"),
        }

        # map dim roles to keys available in the tensor
        keys = {}
        for k in DIM_ROLES:
            for d in dims_map:
                if d.name == k:
                    keys[k] = d.key
                    break
            if k not in keys:
                for d in self.user_dims:
                    if d in DIM_ROLES[k]:
                        keys[k] = d
                        break

        DIM_COMBINATIONS = [
            ["forecast_reference_time", "step"],
            ["forecast_reference_time"],
            ["date", "time", "step"],
            ["date", "time"],
            ["date", "step"],
            ["time", "step"],
            ["step"],
        ]

        for dims in DIM_COMBINATIONS:
            if all(d in keys for d in dims):
                dims_step = [keys[d] for d in dims]
                # use same dim order as in user_dims
                dims = [d for d in self.user_dims if d in dims_step]
                assert len(dims) == len(dims_step), f"Duplicate dims in {dims}"
                other_dims = [d for d in self.user_dims if d not in dims]

                if other_dims:
                    import datetime

                    import numpy as np

                    other_coords = {
                        k: next(iter(self.user_coords[k])) for k in other_dims if k in self.user_coords
                    }

                    vals = np.array(
                        [
                            datetime.datetime.fromisoformat(x)
                            for x in self.source.sel(**other_coords).metadata("valid_datetime")
                        ],
                        dtype=dtype,
                    )

                    shape = tuple([self.user_dims[d] for d in dims])
                    return tuple(dims), vals.reshape(shape)
                else:
                    import datetime

                    import numpy as np

                    vals = np.array(
                        [datetime.datetime.fromisoformat(x) for x in self.source.metadata("valid_datetime")],
                        dtype=dtype,
                    )

                    shape = tuple([self.user_dims[d] for d in dims])
                    return tuple(dims), vals.reshape(shape)
        return None, None

    def __getstate__(self):
        r = {}
        r["source"] = self.source
        r["user_coords"] = self.user_coords
        r["user_shape"] = self.user_shape
        r["user_dims"] = self.user_dims
        r["field_coords"] = self.field_coords
        r["field_shape"] = self.field_shape
        r["field_dims"] = self.field_dims
        r["full_shape"] = self.full_shape
        r["flatten_values"] = self.flatten_values
        return r

    def __setstate__(self, state):
        self.source = state["source"]
        self._user_coords = state["user_coords"]
        self._user_shape = state["user_shape"]
        self._user_dims = state["user_dims"]
        self._field_coords = state["field_coords"]
        self._field_shape = state["field_shape"]
        self._field_dims = state["field_dims"]
        self._full_shape = state["full_shape"]
        self.flatten_values = state["flatten_values"]



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from earthkit.data.utils.metadata.dict import UserMetadata

from . import Source

LOG = logging.getLogger(__name__)


class FieldlistFromDicts(Source):
    def __init__(self, list_of_dicts, *args, **kwargs):
        self.d = list_of_dicts
        self._kwargs = kwargs

    def mutate(self):
        import numpy as np

        from earthkit.data.indexing.fieldlist import SimpleFieldList

        from .array_list import ArrayField

        fields = []
        for f in self.d:
            v = f["values"]
            if isinstance(v, list):
                v = np.array(v)
            fields.append(ArrayField(v, UserMetadata(f, shape=v.shape)))
        return SimpleFieldList(fields=fields)


source = FieldlistFromDicts



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from earthkit.data import from_source

from .multi import MultiSource


class MultiUrl(MultiSource):
    def __init__(
        self, urls, *args, filter=None, merger=None, force=None, lazily=True, sort_urls=False, **kwargs
    ):
        from earthkit.data.utils.url import UrlSpec

        if isinstance(urls, UrlSpec):
            url_spec = urls
        else:
            url_spec = UrlSpec.from_urls(urls, **kwargs)

        assert len(url_spec) > 0

        if sort_urls:
            url_spec = url_spec.sorted()

        sources = [
            from_source(
                "url",
                x,
                filter=filter,
                merger=merger,
                force=force,
                # Load lazily so we can do parallel downloads
                lazily=lazily,
                # **x["kwargs"]
            )
            for x in url_spec
        ]

        super().__init__(sources, filter=filter, merger=merger)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from xr_engine_fixtures import compare_coords  # noqa: E402
from xr_engine_fixtures import compare_dims  # noqa: E402


@pytest.mark.cache
@pytest.mark.parametrize("lazy_load", [True, False])
def test_xr_remapping_1(lazy_load):
    ds0 = from_source(
        "url", earthkit_remote_test_data_file("test-data/xr_engine/level/mixed_pl_ml_small.grib")
    )
    ds = ds0.to_xarray(
        variable_key="_k", remapping={"_k": "{param}_{levelist}_{levtype}"}, lazy_load=lazy_load
    )

    data_vars = ["t_137_ml", "t_500_pl", "t_700_pl", "t_90_ml", "u_137_ml", "u_500_pl", "u_700_pl", "u_90_ml"]
    assert [v for v in ds.data_vars] == data_vars

    dims = {"forecast_reference_time": 4, "step": 2, "latitude": 19, "longitude": 36}
    compare_dims(ds, dims, sizes=True)


@pytest.mark.cache
@pytest.mark.parametrize("lazy_load", [True, False])
@pytest.mark.parametrize(
    "kwargs,coords,dims",
    [
        (
            dict(
                dim_roles={"level": "_k"},
                level_dim_mode="level",
                remapping={"_k": "{levelist}_{levtype}"},
                dim_name_from_role_name=False,
            ),
            {"_k": ["500_pl", "700_pl"]},
            {"forecast_reference_time": 4, "step_timedelta": 2, "_k": 2, "latitude": 19, "longitude": 36},
        ),
        (
            dict(
                dim_roles={"level": "_k"},
                level_dim_mode="level",
                remapping={"_k": "{levelist}_{levtype}"},
                dim_name_from_role_name=True,
            ),
            {"level": ["500_pl", "700_pl"]},
            {"forecast_reference_time": 4, "step": 2, "level": 2, "latitude": 19, "longitude": 36},
        ),
        (
            dict(
                dim_roles={"level": "_k"},
                level_dim_mode="level",
                remapping={"_k": "{levelist}_{levtype}"},
                rename_dims={"level": "_k"},
                dim_name_from_role_name=True,
            ),
            {"_k": ["500_pl", "700_pl"]},
            {"forecast_reference_time": 4, "step": 2, "_k": 2, "latitude": 19, "longitude": 36},
        ),
    ],
)
def test_xr_remapping_2(lazy_load, kwargs, coords, dims):
    ds0 = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl_small.grib"))
    ds = ds0.to_xarray(lazy_load=lazy_load, **kwargs)

    data_vars = ["r", "t"]

    assert [v for v in ds.data_vars] == data_vars

    # coords = {"_k": ["500_pl", "700_pl"]}
    compare_coords(ds, coords)

    # dims = {"forecast_reference_time": 4, "step_timedelta": 2, "_k": 2, "latitude": 19, "longitude": 36}
    compare_dims(ds, dims, sizes=True)
